# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Async process mapset test case admin
"""
import unittest
from flask.json import dumps as json_dumps
from flask.json import loads as json_load
import time

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"

# Module change example for r.slope.aspect with g.region adjustment

process_chain_short_1 = {
    1: {
        "module": "g.region",
        "inputs": {"raster": "elevation@PERMANENT", "res": "1000"},
        "flags": "p",
        "verbose": True,
    },
    2: {
        "module": "r.slope.aspect",
        "inputs": {
            "elevation": "elevation@PERMANENT",
            "format": "degrees",
            "min_slope": "0.0",
        },
        "outputs": {
            "aspect": {"name": "my_aspect_1"},
            "slope": {"name": "my_slope_1"},
        },
        "flags": "a",
        "overwrite": False,
        "verbose": True,
    },
}

process_chain_short_2 = {
    1: {
        "module": "g.region",
        "inputs": {"raster": "elevation@PERMANENT", "res": "1000"},
        "flags": "p",
        "verbose": True,
    },
    2: {
        "module": "r.slope.aspect",
        "inputs": {
            "elevation": "elevation@PERMANENT",
            "format": "degrees",
            "min_slope": "0.0",
        },
        "outputs": {
            "aspect": {"name": "my_aspect_2"},
            "slope": {"name": "my_slope_2"},
        },
        "flags": "a",
        "overwrite": False,
        "verbose": True,
    },
}

test_mapsets = ["Source_A", "Source_B", "Source_C", "Source_D", "Target"]


class AsyncProcessMapsetTestCaseAdmin(ActiniaResourceTestCaseBase):
    def check_remove_test_mapsets(self):
        """
        Check and remove test mapsets that have been created in the test-run
        Returns:

        """

        for mapset in test_mapsets:
            # Unlock mapset for deletion
            rv = self.server.delete(
                URL_PREFIX
                + "/locations/%s/mapsets/%s/lock" % ("nc_spm_08", mapset),
                headers=self.admin_auth_header,
            )
            print(rv.data)

        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        mapsets = json_load(rv.data)["process_results"]

        for mapset in test_mapsets:
            if mapset in mapsets:
                # Delete the mapset if it already exists
                rv = self.server.delete(
                    URL_PREFIX + "/locations/nc_spm_08/mapsets/%s" % mapset,
                    headers=self.admin_auth_header,
                )
                print(rv.data)
                self.assertEqual(
                    rv.status_code,
                    200,
                    "HTML status code is wrong %i %s"
                    % (rv.status_code, mapset),
                )
                self.assertEqual(
                    rv.mimetype,
                    "application/json",
                    "Wrong mimetype %s" % rv.mimetype,
                )

    def test_1_merge_no_access_to_target_mapset_error(self):
        """No access to target mapset error test"""
        # Try merge source mapsets into target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/user1/merging_async",
            headers=self.user_auth_header,
            data=json_dumps(["Source_A", "Source_B"]),
            content_type="application/json",
        )

        print(rv.data)
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.user_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError",
        )

    def test_2_merge_missing_target_mapset_error(self):
        """Missing target mapset test"""
        self.check_remove_test_mapsets()

        # Try merge source mapsets into target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target/merging_async",
            headers=self.admin_auth_header,
            data=json_dumps(["Source_A", "Source_B"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError",
        )

    def test_3_merge_missing_source_mapsets_error(self):
        """Test error for missing source mapsets"""
        self.check_remove_test_mapsets()

        # Create target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Try merge source mapsets into target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target/merging_async",
            headers=self.admin_auth_header,
            data=json_dumps(["Source_A", "Source_B"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError",
        )

    def test_4_merge_empty_mapset_list(self):
        """Test error for missing source mapsets"""
        self.check_remove_test_mapsets()

        # Create target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Try merge source mapsets into target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target/merging_async",
            headers=self.admin_auth_header,
            data=json_dumps([]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError",
        )

    def test_5_merge_two_mapsets(self):
        """Test the merging of two mapsets into a target mapset"""
        self.check_remove_test_mapsets()

        # Create the source mapsets
        rv = self.server.post(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Source_A/processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_short_1),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

        rv = self.server.post(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Source_B/processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_short_2),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

        # Create target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Merge source mapsets into target mapset
        rv = self.server.post(
            URL_PREFIX + "/locations/nc_spm_08/mapsets/Target/merging_async",
            headers=self.admin_auth_header,
            data=json_dumps(["Source_A", "Source_B"]),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(rv, headers=self.admin_auth_header)

        # Check copied raster
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Target/raster_layers/my_aspect_1",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertTrue(
            "my_aspect_1", json_load(rv.data)["process_results"]["map"]
        )
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Target/raster_layers/my_aspect_2",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertTrue(
            "my_aspect_2", json_load(rv.data)["process_results"]["map"]
        )
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Target/raster_layers/my_slope_1",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertTrue(
            "my_slope_1", json_load(rv.data)["process_results"]["map"]
        )
        rv = self.server.get(
            URL_PREFIX
            + "/locations/nc_spm_08/mapsets/Target/raster_layers/my_slope_2",
            headers=self.admin_auth_header,
        )
        print(rv.data)
        self.assertTrue(
            "my_slope_2", json_load(rv.data)["process_results"]["map"]
        )

        time.sleep(1)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Actinia user test case
"""
from flask.json import loads as json_load
from werkzeug.datastructures import Headers
import unittest
import base64

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class ActiniaUserTestCase(ActiniaResourceTestCaseBase):
    """
    Test the REST User credentials, creation, deletion API

    User can only be received, created and deleted
    as user with an admin role.
    """

    created_users = list()

    def tearDown(self):
        # delete user
        rv = self.server.get(
            URL_PREFIX + "/users", headers=self.admin_auth_header
        )
        self.assertEqual(rv.status_code, 200, "Getting user list failed!")
        user_list = json_load(rv.data)["user_list"]
        for user_id in set(self.created_users):
            if user_id in user_list:
                delete_user_url = f"{URL_PREFIX}/users/{user_id}"
                rv2 = self.server.delete(
                    delete_user_url, headers=self.admin_auth_header
                )
                self.assertEqual(
                    rv2.status_code,
                    200,
                    "Deleting of user %s does not work" % user_id,
                )
        super(ActiniaUserTestCase, self).tearDown()

    def test_list_users(self):
        """
        List all users in the database
        :return:
        """
        rv = self.server.get(
            URL_PREFIX + "/users", headers=self.admin_auth_header
        )
        print("GET /users", rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        users = ""
        if "guest" in json_load(rv.data)["user_list"]:
            users = "guest"

        self.assertEqual(users, "guest", "Wrong user listed")

    def test_create_get_delete_user(self):
        """
        Create a new user, access it as admin and as normal user and delete it
        :return:
        """

        # Create a new user
        new_user_id = "thomas"
        new_group = "test"
        new_password = "12345"
        self.created_users.append(new_user_id)

        create_user_url = (
            f"{URL_PREFIX}/users/{new_user_id}?"
            f"password={new_password}&group={new_group}"
        )
        rv = self.server.post(create_user_url, headers=self.admin_auth_header)
        self.assertEqual(
            rv.status_code,
            201,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Try to create the user again which should fail
        rv = self.server.post(create_user_url, headers=self.admin_auth_header)
        self.assertEqual(
            rv.status_code, 400, f"HTML status code is wrong {rv.status_code}"
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Get the new user entry
        rv = self.server.get(
            URL_PREFIX + f"/users/{new_user_id}",
            headers=self.admin_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )
        self.assertEqual(json_load(rv.data)["user_id"], new_user_id)
        self.assertEqual(json_load(rv.data)["user_role"], "user")

        # Get the admin user entry
        rv = self.server.get(
            URL_PREFIX + "/users/admin", headers=self.admin_auth_header
        )
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )
        self.assertEqual(json_load(rv.data)["user_id"], "admin")
        self.assertEqual(json_load(rv.data)["user_role"], "admin")

        # Delete the user as admin
        rv = self.server.delete(
            URL_PREFIX + "/users/%s" % new_user_id,
            headers=self.admin_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_create_get_delete_user_unprivileged(self):
        """
        Create, get and delete as normal user
        :return:
        """

        # Create a new user
        new_user_id = "thomas"
        new_group = "test"
        new_password = "12345"
        self.created_users.append(new_user_id)

        # Now try to create a new user with the thomas account which should
        # fail!
        new_auth_header = Headers()
        auth = bytes("%s:%s" % (new_user_id, new_password), "utf-8")
        new_auth_header.add(
            "Authorization", "Basic " + base64.b64encode(auth).decode()
        )

        rv = self.server.post(
            URL_PREFIX
            + "/users/%s?password=%s&group=%s"
            % (new_user_id, new_password, new_group),
            headers=new_auth_header,
        )
        print("POST /users/thomas", rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )

        # Try to get the user credentials as non-admin which should fail!
        rv = self.server.get(
            URL_PREFIX + "/users/%s" % new_user_id, headers=new_auth_header
        )
        print("GET /users/thomas", rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )

        # Try to delete the new user as non-admin which should fail!
        rv = self.server.delete(
            URL_PREFIX + "/users/%s" % new_user_id, headers=new_auth_header
        )
        print("DELETE /users/thomas", rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_get_own_userid(self):
        """
        Also users and guests should be allowed to request their own user.
        :return:
        """
        # user requests other user (should fail)
        rv = self.server.get(
            URL_PREFIX + "/users/admin", headers=self.user_auth_header
        )
        self.assertEqual(
            rv.status_code,
            401,
            f"HTML status code is wrong {rv.status_code} for get other user",
        )
        # user requests own user
        rv = self.server.get(
            URL_PREFIX + "/users/user", headers=self.user_auth_header
        )
        self.assertEqual(
            rv.status_code,
            200,
            f"HTML status code is wrong {rv.status_code} for get user from a "
            "user.",
        )
        # guest requests other user (should fail)
        rv = self.server.get(
            URL_PREFIX + "/users/user", headers=self.guest_auth_header
        )
        self.assertEqual(
            rv.status_code,
            401,
            f"HTML status code is wrong {rv.status_code} for get other user",
        )
        # guest requests own user
        rv = self.server.get(
            URL_PREFIX + "/users/guest", headers=self.guest_auth_header
        )
        self.assertEqual(
            rv.status_code,
            200,
            f"HTML status code is wrong {rv.status_code} for get user from a "
            "guest.",
        )
        # admin requests other user
        rv = self.server.get(
            URL_PREFIX + "/users/user", headers=self.admin_auth_header
        )
        self.assertEqual(
            rv.status_code,
            200,
            f"HTML status code is wrong {rv.status_code} for get user from a "
            "admin.",
        )

    def test_token_generation_admin(self):
        rv = self.server.get(
            URL_PREFIX + "/token", headers=self.admin_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_api_key_generation_admin(self):
        rv = self.server.get(
            URL_PREFIX + "/api_key", headers=self.admin_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_user(self):
        rv = self.server.get(
            URL_PREFIX + "/token", headers=self.user_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_user_expire(self):
        rv = self.server.get(
            URL_PREFIX + "/token?expiration_time=3600",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_guest(self):
        rv = self.server.get(
            URL_PREFIX + "/token", headers=self.guest_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_wrong_no_auth(self):
        rv = self.server.get(URL_PREFIX + "/token")
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "text/html", "Wrong mimetype %s" % rv.mimetype
        )

    def test_api_key_generation_wrong_user(self):
        rv = self.server.get(
            URL_PREFIX + "/api_key", headers=self.user_auth_header
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            401,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_user_expire_error(self):
        rv = self.server.get(
            URL_PREFIX + "/token?expiration_time=1000000000",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_token_generation_user_expire_wrong(self):
        rv = self.server.get(
            URL_PREFIX + "/token?expiration_time=blablub",
            headers=self.user_auth_header,
        )
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            400,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Base class for asynchronous and synchronous responses
"""
import os
import pickle
import time
import uuid
from datetime import datetime
from flask import make_response, jsonify
from flask import request, g
from flask.json import loads as json_loads
from flask_restful_swagger_2 import Resource
from actinia_core.rest.base.user_auth import check_user_permissions
from actinia_core.rest.base.user_auth import create_dummy_user
from actinia_core.core.common.app import auth
from actinia_core.core.common.app import flask_api
from actinia_core.core.common.config import global_config
from actinia_core.core.common.api_logger import log_api_call
from actinia_core.core.messages_logger import MessageLogger
from actinia_core.core.resources_logger import ResourceLogger
from actinia_core.core.resource_data_container import ResourceDataContainer
from actinia_core.models.response_models import ProcessingResponseModel
from actinia_core.models.response_models import (
    create_response_from_model,
    ApiInfoModel,
)
from actinia_core.rest.resource_streamer import RequestStreamerResource
from actinia_core.rest.resource_management import ResourceManager

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ResourceBase(Resource):
    """
    This is the base class for all asynchronous and synchronous processing
    resources.
    """

    # Authorization is required for all resources
    # User permissions must be checked for all resources
    # API logging is required for all resources
    # decorators = [log_api_call, check_user_permissions, auth.login_required]

    decorators = []

    if global_config.LOG_API_CALL is True:
        decorators.append(log_api_call)

    if global_config.CHECK_CREDENTIALS is True:
        decorators.append(check_user_permissions)

    if global_config.LOGIN_REQUIRED is True:
        decorators.append(auth.login_required)
    else:
        decorators.append(create_dummy_user)

    def __init__(self, resource_id=None, iteration=None, post_url=None):
        # Configuration
        Resource.__init__(self)

        # Store the user id, user group and all credentials of the current user
        self.user = g.user
        self.user_id = g.user.get_id()
        self.user_group = g.user.get_group()
        self.user_role = g.user.get_role()
        self.has_superadmin_role = g.user.has_superadmin_role()
        self.user_credentials = g.user.get_credentials()

        self.orig_time = time.time()
        self.orig_datetime = str(datetime.now())

        kwargs = dict()
        kwargs["host"] = global_config.REDIS_SERVER_URL
        kwargs["port"] = global_config.REDIS_SERVER_PORT
        if (
            global_config.REDIS_SERVER_PW
            and global_config.REDIS_SERVER_PW is not None
        ):
            kwargs["password"] = global_config.REDIS_SERVER_PW
        self.resource_logger = ResourceLogger(**kwargs)
        del kwargs

        self.message_logger = MessageLogger()

        self.grass_data_base = global_config.GRASS_DATABASE
        self.grass_user_data_base = global_config.GRASS_USER_DATABASE
        self.grass_base_dir = global_config.GRASS_GIS_BASE
        self.grass_start_script = global_config.GRASS_GIS_START_SCRIPT
        self.grass_addon_path = global_config.GRASS_ADDON_PATH
        self.download_cache = os.path.join(
            global_config.DOWNLOAD_CACHE, self.user_id
        )

        # Set the resource id
        if resource_id is None:
            # Generate the resource id
            self.request_id, self.resource_id = self.generate_uuids()
        else:
            self.resource_id = resource_id
            self.request_id = self.generate_request_id_from_resource_id()

        if global_config.QUEUE_TYPE == "per_job":
            self.queue = "%s_%s" % (
                global_config.WORKER_QUEUE_PREFIX,
                self.resource_id,
            )
        elif global_config.QUEUE_TYPE == "per_user":
            self.queue = "%s_%s" % (
                global_config.WORKER_QUEUE_PREFIX,
                self.user_id,
            )
        elif global_config.QUEUE_TYPE == "redis":
            self.queue = "%s_%s" % (global_config.WORKER_QUEUE_PREFIX, "count")
        else:
            self.queue = "local"

        # set iteration and post_url
        self.iteration = iteration
        self.post_url = post_url

        # The base URL's for resources that will be streamed
        self.resource_url_base = None

        # Generate the status URL
        self.status_url = flask_api.url_for(
            ResourceManager,
            user_id=self.user_id,
            resource_id=self.resource_id,
            _external=True,
        )

        if (
            global_config.FORCE_HTTPS_URLS is True
            and "http://" in self.status_url
        ):
            self.status_url = self.status_url.replace("http://", "https://")

        self.request_url = request.url
        self.resource_url = None
        self.request_data = None
        self.response_data = None
        self.job_timeout = 0

        # Replace this with the correct response model in subclasses
        # The class that is used to create the response
        self.response_model_class = ProcessingResponseModel

        # Put API information in the response for later accounting
        kwargs = {
            "endpoint": request.endpoint,
            "method": request.method,
            "path": request.path,
            "request_url": self.request_url,
        }
        if self.post_url is not None:
            kwargs["post_url"] = self.post_url
        self.api_info = ApiInfoModel(**kwargs)

    def create_error_response(self, message, status="error", http_code=400):
        """Create an error response, that by default sets the status to error
        and the http_code to 400

        This method sets the self.response_data variable.

        Args:
            message: The error message
            status: The status, by default error
            http_code: The hhtp code by default 400

        """
        self.response_data = create_response_from_model(
            self.response_model_class,
            status=status,
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.queue,
            iteration=self.iteration,
            process_log=None,
            results={},
            message=message,
            http_code=http_code,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            status_url=self.status_url,
            api_info=self.api_info,
        )

    def get_error_response(self, message, status="error", http_code=400):
        """Return the error response.

        This function will generate an error response using make_response()
        In addition, a resource update is send using the error response.

        Args:
            message: The error message
            status: The status, by default error
            http_code: The http code by default 400

        Returns:
            the result of make_response()

        """
        self.create_error_response(
            message=message, status=status, http_code=http_code
        )
        self.resource_logger.commit(
            user_id=self.user_id,
            resource_id=self.resource_id,
            iteration=self.iteration,
            document=self.response_data,
        )
        http_code, response_model = pickle.loads(self.response_data)
        return make_response(jsonify(response_model), http_code)

    def check_for_json(self):
        """Check if the Payload is a JSON document

        Return: bool:
            True in case of success, False otherwise
        """
        # First check for the data field and create JSON from it
        if hasattr(request, "data") is True:
            try:
                self.request_data = json_loads(request.data)
            except Exception as e:
                self.create_error_response(
                    message="No JSON data in request: Exception: %s" % str(e)
                )
                return False

        if request.is_json is False:
            self.create_error_response(message="No JSON data in request")
            return False

        self.request_data = request.get_json()

        return True

    def check_for_xml(self):
        """Check if the Payload is a XML document

        Return: bool:
            True in case of success, False otherwise
        """

        if "Content-Type" not in request.headers:
            self.create_error_response(message="XML content type is required.")
            return False

        if "XML" not in request.headers["Content-Type"].upper():
            self.create_error_response(message="XML content type is required.")
            return False

        # Check if payload was provided
        if hasattr(request, "data") is False:
            self.create_error_response(
                message="No XML data section in HTTP header."
            )
            return False

        if request.data:
            self.request_data = request.data
            return True
        else:
            self.create_error_response(
                message="Empty XML data section in HTTP header."
            )
            return False

    def preprocess(
        self,
        has_json=True,
        has_xml=False,
        location_name=None,
        mapset_name=None,
        map_name=None,
        process_chain_list=None,
    ):
        """Preprocessing steps for asynchronous processing

            - Check if the request has a data field
            - Check if the module chain description can be loaded
            - Initialize the response and request ids as well as the
              url for status polls
            - Send an accept entry to the resource redis database

        Args:
            has_json (bool): Set True if the request has JSON data, False
                             otherwise
            has_xml (bool): Set True if the request has XML data, False
                            otherwise
            location_name (str): The name of the location to work in
            mapset_name (str): The name of the target mapset in which the
                               computation should be performed
            map_name: The name of the map or other resource (raster, vector,
                      STRDS, color, ...)
            process_chain_list (dict): The process chain list (e.g. for the
                                       job resumption when no new postbody is
                                       send in the PUT request)

        Returns:
            The ResourceDataContainer that contains all required information
            for the async process or None if the request was wrong. Then use
            the self.response_data variable to send a response.

        """
        if has_json is True and has_xml is True:
            if request.is_json is True:
                self.request_data = request.get_json()
            else:
                if self.check_for_xml() is False:
                    return None
        elif has_xml is True:
            if self.check_for_xml() is False:
                return None
        elif has_json is True:
            if self.check_for_json() is False:
                return None
        elif process_chain_list is not None:
            self.request_data = process_chain_list

        # Compute the job timeout of the worker queue from the user credentials
        process_time_limit = self.user_credentials["permissions"][
            "process_time_limit"
        ]
        process_num_limit = self.user_credentials["permissions"][
            "process_num_limit"
        ]
        self.job_timeout = int(process_time_limit * process_num_limit * 20)

        # Create the resource URL base and use a placeholder for the file name
        # The placeholder __None__ must be replaced by the resource URL
        # generator
        self.resource_url_base = flask_api.url_for(
            RequestStreamerResource,
            user_id=self.user_id,
            resource_id=self.resource_id,
            file_name="__None__",
            _external=True,
        )

        if (
            global_config.FORCE_HTTPS_URLS is True
            and "http://" in self.resource_url_base
        ):
            self.resource_url_base = self.resource_url_base.replace(
                "http://", "https://"
            )

        # Create the accepted response that will be always send
        self.response_data = create_response_from_model(
            self.response_model_class,
            status="accepted",
            user_id=self.user_id,
            resource_id=self.resource_id,
            queue=self.queue,
            iteration=self.iteration,
            process_log=None,
            results={},
            message="Resource accepted",
            http_code=200,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            status_url=self.status_url,
            api_info=self.api_info,
        )

        # Send the status to the database
        self.resource_logger.commit(
            self.user_id, self.resource_id, self.iteration, self.response_data
        )

        # Return the ResourceDataContainer that includes all
        # required data for the asynchronous processing
        return ResourceDataContainer(
            grass_data_base=self.grass_data_base,
            grass_user_data_base=self.grass_user_data_base,
            grass_base_dir=self.grass_base_dir,
            request_data=self.request_data,
            user_id=self.user_id,
            user_group=self.user_group,
            user_credentials=self.user_credentials,
            resource_id=self.resource_id,
            iteration=self.iteration,
            status_url=self.status_url,
            api_info=self.api_info,
            resource_url_base=self.resource_url_base,
            orig_time=self.orig_time,
            orig_datetime=self.orig_datetime,
            config=global_config,
            location_name=location_name,
            mapset_name=mapset_name,
            map_name=map_name,
        )

    def generate_uuids(self):
        """Return a unique request and resource id based on uuid4

        Returns:
            (str, str):
            request_id, resource_id as strings
        """
        new_uuid = uuid.uuid4()
        request_id = "request_id-" + str(new_uuid)
        resource_id = "resource_id-" + str(new_uuid)

        return request_id, resource_id

    def generate_request_id_from_resource_id(self):
        return self.resource_id.replace("resource_id-", "request_id-")

    def wait_until_finish(self, poll_time=0.2):
        """Wait until a resource finished, terminated or failed with an error

        Call this method if a job was enqueued and the POST/GET/DELETE/PUT
        method should wait for it

        Args:
            poll_time (float): Time to sleep between Redis db polls for process
                               status requests

        Returns:
            (int, dict)
            The http_code and the generated data dictionary
        """
        # Wait for the async process by asking the redis database for updates
        while True:
            response_data = self.resource_logger.get(
                self.user_id, self.resource_id, self.iteration
            )
            if not response_data:
                message = (
                    "Unable to receive process status. User id "
                    "%s resource id %s and iteration %d"
                    % (self.user_id, self.resource_id, self.iteration)
                )
                return make_response(message, 400)

            http_code, response_model = pickle.loads(response_data)
            if (
                response_model["status"] == "finished"
                or response_model["status"] == "error"
                or response_model["status"] == "timeout"
                or response_model["status"] == "terminated"
            ):
                break
            time.sleep(poll_time)

        return (http_code, response_model)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Redis connection interface
"""
import rq
from redis import Redis
from actinia_core.core.redis_user import redis_user_interface
from actinia_core.core.redis_api_log import redis_api_log_interface
from actinia_core.core.logging_interface import log
from .config import global_config
from .process_queue import enqueue_job as enqueue_job_local

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"

job_queues = []
redis_conn = None


def connect(host, port, pw=None):
    """Connect all required redis interfaces that should be used
       in the main server process.

       These interfaces are connected here for performance reasons.
       The redis job queue is initialized here as well.

    Args:
        host (str): The hostname of the redis server
        port (str): The port of the redis server
        pw (str): The password of the redis server

    """
    redis_user_interface.connect(host, port, pw)
    redis_api_log_interface.connect(host, port, pw)


def disconnect():
    """Disconnect all required redis interfaces"""
    redis_user_interface.disconnect()
    redis_api_log_interface.disconnect()


def __create_job_queue(queue_name):
    """Create a single job queue for asynchronous processing

    Args:
        queue_name: The name of the queue

    """
    # Redis work queue and connection
    global job_queues, redis_conn

    if not any(q.name == queue_name for q in job_queues):
        host = global_config.REDIS_QUEUE_SERVER_URL
        port = global_config.REDIS_QUEUE_SERVER_PORT
        password = global_config.REDIS_QUEUE_SERVER_PASSWORD

        kwargs = dict()
        kwargs["host"] = host
        kwargs["port"] = port
        if password and password is not None:
            kwargs["password"] = password
        redis_conn = Redis(**kwargs)

        string = "Create queue %s with server %s:%s" % (queue_name, host, port)
        log.info(string)
        queue = rq.Queue(queue_name, connection=redis_conn)
        job_queues.append(queue)


def __enqueue_job_redis(queue, timeout, func, *args):
    """Enqueue a job in the job queues

    Args:
        func: The function to call from the subprocess
        *args: The function arguments
    """

    log.info("Enqueue job in queue %s" % queue.name)
    # Below timeout is defined in resource_base.pyL295.
    # If it is higher than 2147483647, it will be set to never expire.
    # Else it would raise an error:
    # int(process_time_limit * process_num_limit * 20)
    # which is 630720000000 and raises in worker:
    # OverflowError: Python int too large to convert to C int
    if timeout > 2147483647:
        timeout = -1  # never exprire
    ret = queue.enqueue(
        func,
        *args,
        job_timeout=timeout,
        ttl=global_config.REDIS_QUEUE_JOB_TTL,
        result_ttl=global_config.REDIS_QUEUE_JOB_TTL,
    )
    log.info(ret)


def enqueue_job(timeout, func, *args, queue_type_overwrite=None):
    """Write the provided function in a queue

    Args:
        timeout: The timeout of the process
        func: The function to call from the subprocess/worker
        *args: The function arguments
    """
    global job_queues, redis_conn
    num_queues = global_config.NUMBER_OF_WORKERS
    queue_type = global_config.QUEUE_TYPE
    queue_name = "local"
    if queue_type_overwrite:
        queue_type = global_config.QUEUE_TYPE_OVERWRITE

    if queue_type == "per_job":
        resource_id = args[0].resource_id
        queue_name = "%s_%s" % (global_config.WORKER_QUEUE_PREFIX, resource_id)
        __create_job_queue(queue_name)
        for i in job_queues:
            if i.name == queue_name:
                args[0].set_queue_name(queue_name)
                __enqueue_job_redis(i, timeout, func, *args)

    elif queue_type == "per_user":
        user_id = args[0].user_id
        queue_name = "%s_%s" % (global_config.WORKER_QUEUE_PREFIX, user_id)
        # Run __create_job_queue everytime.
        # If queue already exists, it does nothing.
        __create_job_queue(queue_name)
        for i in job_queues:
            if i.name == queue_name:
                args[0].set_queue_name(queue_name)
                __enqueue_job_redis(i, timeout, func, *args)

    elif queue_type == "redis":
        if job_queues == []:
            for i in range(num_queues):
                queue_name = "%s_%s" % (global_config.WORKER_QUEUE_PREFIX, i)
                __create_job_queue(queue_name)
        # The redis incr approach is used here
        # to chose for each job a different queue
        num = redis_conn.incr("actinia_worker_count", 1)
        current_queue = num % num_queues
        args[0].set_queue_name(job_queues[current_queue].name)
        __enqueue_job_redis(job_queues[current_queue], timeout, func, *args)

    elif queue_type == "local":
        # __enqueue_job_local(timeout, func, *args)
        args[0].set_queue_name(queue_name)
        enqueue_job_local(timeout, func, *args)
        return
        # Just in case the current process queue does not work
        # Then use the most simple solution by just starting the process
        from multiprocessing import Process

        p = Process(target=func, args=args)
        p.start()



#!flask/bin/python
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Actinia server script

This script is for testing purpose only, since the actinia server should
run behind the uwsgi application server.
"""
import argparse
import os
from actinia_core.endpoints import create_endpoints
from actinia_core.core.common.app import flask_app
from actinia_core.core.common.config import global_config, DEFAULT_CONFIG_PATH
from actinia_core.core.common.redis_interface import connect
from actinia_core.core.common.process_queue import create_process_queue

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


def main():
    parser = argparse.ArgumentParser(
        description="Start actinia server. A running redis server"
        + " is required."
    )

    parser.add_argument(
        "--host",
        type=str,
        required=False,
        default="0.0.0.0",
        help="The IP address that should be used for the server",
    )

    parser.add_argument(
        "--port",
        type=int,
        required=False,
        default=5000,
        help="The port that should be used for the server",
    )

    parser.add_argument(
        "--debug",
        type=bool,
        required=False,
        default=True,
        help="Set True to activate debugging",
    )

    args = parser.parse_args()

    if os.path.exists(DEFAULT_CONFIG_PATH) is True and os.path.isfile(
        DEFAULT_CONFIG_PATH
    ):
        global_config.read(DEFAULT_CONFIG_PATH)

    # Create the endpoints based on the global config
    create_endpoints()

    # Connect the redis interfaces
    redis_args = (
        global_config.REDIS_SERVER_URL,
        global_config.REDIS_SERVER_PORT,
    )
    if (
        global_config.REDIS_SERVER_PW
        and global_config.REDIS_SERVER_PW is not None
    ):
        redis_args = (*redis_args, global_config.REDIS_SERVER_PW)

    connect(*redis_args)
    del redis_args

    # Create the process queue
    create_process_queue(global_config)

    flask_app.run(host=args.host, port=args.port)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Actinia Core Exceptions that should be used in case an error occurs that is
related to the Actinia Core functionality
"""

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class AsyncProcessError(Exception):
    """
    Raise this exception in case the asynchronous processing faces an error
    """

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)


class RsyncError(Exception):
    """Raise this exception in case the rsync of the interim result fails"""

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)


class AsyncProcessTermination(Exception):
    """
    Raise this exception in case the termination requests was executed
    successfully
    """

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)


class AsyncProcessTimeLimit(Exception):
    """Raise this exception in case the process time limit was reached"""

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)


class GoogleCloudAPIError(Exception):
    """Raise this exception in case something went wrong in
    when accessing the google API
    """

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)


class SecurityError(Exception):
    """Raise this exception in case some security problem occurs"""

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Mapset management resources

* List all mapsets
* Create mapset, Delete mapset, Get info about a mapset
* Lock mapset, unlock mapset, get mapset lock status
"""

import shutil
from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.common.exceptions import AsyncProcessError
from actinia_core.models.response_models import (
    StringListProcessingResultResponseModel,
    MapsetInfoResponseModel,
    RegionModel,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class PersistentMapsetLister(PersistentProcessing):
    """List all mapsets in a location"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.response_model_class = StringListProcessingResultResponseModel

    def _execute(self):
        self._setup()
        # Create the temporary database and link all available mapsets into is
        self._create_temp_database()

        pc = {
            "1": {
                "module": "g.mapsets",
                "inputs": {"separator": "newline"},
                "flags": "l",
            }
        }

        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base, mapset_name="PERMANENT"
        )

        self._execute_process_list(process_list)

        mapset_lists = []
        mapsets = self.module_output_log[0]["stdout"].split()

        for mapset in mapsets:
            mapset_lists.append(mapset.strip())

        self.module_results = mapset_lists


class PersistentGetProjectionRegionInfo(PersistentProcessing):
    """Read the current region and projection information"""

    integer_values = [
        "projection",
        "zone",
        "rows",
        "rows3",
        "cols",
        "cols3",
        "depths",
        "cells",
        "cells3",
    ]
    float_values = [
        "n",
        "s",
        "e",
        "w",
        "t",
        "b",
        "nsres",
        "nsres3",
        "ewres",
        "ewres3",
        "tbres",
    ]

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.response_model_class = MapsetInfoResponseModel

    def _execute(self):
        self._setup()

        pc = {
            "1": {"module": "g.region", "flags": "ug3"},
            "2": {"module": "g.proj", "flags": "fw"},
        }

        # Do not check the region size
        self.skip_region_check = True
        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )
        self._create_temp_database(self.required_mapsets)
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(process_list)

        mapset_region = {}
        region_settings = self.module_output_log[0]["stdout"].split()

        for region_token in region_settings:
            if "=" in region_token:
                key, value = region_token.strip().split("=")
                if key in self.integer_values:
                    mapset_region[key] = int(value)
                if key in self.float_values:
                    mapset_region[key] = float(value)

        self.module_results = dict(
            region=RegionModel(**mapset_region),
            projection=self.module_output_log[1]["stdout"],
        )

        # self.module_results = MapsetInfoModel(
        #     region=RegionModel(**mapset_region),
        #     projection=self.module_output_log[1]["stdout"]
        # )


class PersistentMapsetCreator(PersistentProcessing):
    """Create a mapset in an existing location"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        # Create temporary database
        self._create_temp_database()

        pc = {"1": {"module": "g.mapsets", "flags": "l"}}

        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base, mapset_name="PERMANENT"
        )

        self._execute_process_list(process_list)

        mapset_list = []
        mapsets = self.module_output_log[0]["stdout"].split()

        for mapset in mapsets:
            mapset_list.append(mapset.strip())

        if self.target_mapset_name in mapset_list:
            raise AsyncProcessError(
                "Mapset <%s> exists." % self.target_mapset_name
            )

        # Create the new temporary mapset and merge it into the user database
        # location
        self._check_lock_target_mapset()
        self.required_mapsets = ["PERMANENT"]
        self._create_temporary_mapset(temp_mapset_name=self.temp_mapset_name)
        self._copy_merge_tmp_mapset_to_target_mapset()

        self.finish_message = (
            "Mapset <%s> successfully created." % self.target_mapset_name
        )


class PersistentMapsetDeleter(PersistentProcessing):
    """Delete a mapset from a location

    1. Create temporary database
    2. Check if PERMANENT mapset or global mapset which are not allowed to be
       deleted
    3. Check if the mapset exists
    """

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()

        # For debug purpose
        # self.lock_interface.unlock(self.target_mapset_lock_id)

        if "PERMANENT" == self.target_mapset_name:
            raise AsyncProcessError(
                "The PERMANENT mapset can not be deleted. "
                "You must remove the location to get rid of it."
            )

        # Delete existing mapset
        self._check_lock_target_mapset()
        # The variable self.orig_mapset_path is set by
        # _check_lock_target_mapset()
        if self.target_mapset_exists is True:
            shutil.rmtree(self.orig_mapset_path)
            self.lock_interface.unlock(self.target_mapset_lock_id)
            self.finish_message = (
                "Mapset <%s> successfully removed." % self.target_mapset_name
            )
        else:
            raise AsyncProcessError(
                "Mapset <%s> does not exits" % self.target_mapset_name
            )


class PersistentGetMapsetLock(PersistentProcessing):
    """Get the mapset lock status"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        self._check_target_mapset_exists()
        if self.target_mapset_exists is False:
            raise AsyncProcessError(
                f"Unable to get lock status of mapset <{self.mapset_name}>"
                f" in location <{self.location_name}>: Mapset does not "
                "exist"
            )
        else:
            self.module_results = self.lock_interface.get(
                self.target_mapset_lock_id
            )
            self.finish_message = "Mapset lock state: %s" % str(
                self.module_results
            )


class PersistentMapsetLocker(PersistentProcessing):
    """Lock a mapset"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        self._check_lock_target_mapset()
        if self.target_mapset_exists is False:
            raise AsyncProcessError(
                "Unable to lock mapset <%s>. Mapset doesn not exists."
                % self.target_mapset_name
            )

        self.finish_message = (
            "Mapset <%s> successfully locked" % self.target_mapset_name
        )

    def _final_cleanup(self):
        """
        Final cleanup called in the run function at the very end of processing
        """
        # Clean up and remove the temporary files
        self._cleanup()


class PersistentMapsetUnlocker(PersistentProcessing):
    """Unlock a locked mapset"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        self._check_target_mapset_exists()
        if self.target_mapset_exists is False:
            raise AsyncProcessError(
                (
                    "Unable to unlock mapset <%s> in location <%s>:"
                    " Mapset does not exist"
                )
                % (self.mapset_name, self.location_name)
            )
        else:
            self.lock_interface.unlock(self.target_mapset_lock_id)
            self.finish_message = (
                "Mapset <%s> successfully unlocked" % self.target_mapset_name
            )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Redis server user interface
"""

from actinia_core.core.common.redis_base import RedisBaseInterface
import pickle

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class RedisUserInterface(RedisBaseInterface):
    """The Redis user database interface

    A single user is stored as Hash with several entries:
        - User id aka user name that must be unique
        - User role (superadmin, admin, user, guest)
        - User group
        - password hash
        - Permission dictionary

    In addition is the user_id saved in a hash that contains all user ids.
    """

    # We use two databases The user ID and the User name database
    # The user ID and user name databases are hashes
    user_id_hash_prefix = "USER-ID-HASH-PREFIX::"
    user_id_db = "USER-ID-DATABASE"

    def __init__(self):
        RedisBaseInterface.__init__(self)

    def get_password_hash(self, user_id):
        """Return the password hash of the user_id

        HGET User-id password_hash

        Args:
            user_id (str): The user id

        Returns:
             str:
             The password hash of the user id
        """
        return self.redis_server.hget(
            self.user_id_hash_prefix + user_id, "password_hash"
        ).decode()

    def get_role(self, user_id):
        """Return the role of the user

        HGET User-id user_role

        Args:
            user_id: The user id

        Returns:
             str:
             The api key of the user id
        """
        return self.redis_server.hget(
            self.user_id_hash_prefix + user_id, "user_role"
        ).decode()

    def get_group(self, user_id):
        """Return the group of the user

        HGET User-id user_group

        Args:
            user_id: The user id

        Returns:
             str:
             The user group
        """
        return self.redis_server.hget(
            self.user_id_hash_prefix + user_id, "user_group"
        ).decode()

    def get_credentials(self, user_id):
        """Return a dictionary that contains the user credentials

        HGETALL User-id db

        Args:
            user_id: The user id

        Returns:
            dict:
            A dictionary that contains the user credentials
        """
        creds = {}
        user_creds = self.redis_server.hgetall(
            self.user_id_hash_prefix + user_id
        )

        if user_creds:
            creds["user_id"] = user_creds[b"user_id"].decode()
            creds["password_hash"] = user_creds[b"password_hash"].decode()
            creds["user_role"] = user_creds[b"user_role"].decode()
            creds["user_group"] = user_creds[b"user_group"].decode()
            creds["permissions"] = pickle.loads(user_creds[b"permissions"])
        return creds

    def add(self, user_id, user_group, password_hash, user_role, permissions):
        """
        Add a user to the user database

        HSET User-id db -> add user-id
        HSET User-id-hash db -> add a new user with all required entries

        Args:
            user_id (str): The user id
            user_group (str): The user group
            password_hash (str): The password hash
            user_role (str): Get the role of the user ("admin", "user",
                             "guest")
            permissions (dict): A dictionary of permissions

        Returns:
            bool:
            True is success, False if user is already in the database
        """
        if (
            self.redis_server.exists(self.user_id_hash_prefix + user_id)
            is True
        ):
            return False
        pstring = pickle.dumps(permissions)

        lock = self.redis_server.lock(name="add_user_lock", timeout=1)
        lock.acquire()
        # First add the user-id to the user id database
        self.redis_server.hset(self.user_id_db, user_id, user_id)

        mapping = {
            "user_id": user_id,
            "password_hash": password_hash,
            "user_role": user_role,
            "user_group": user_group,
            "permissions": pstring,
        }
        # Make the database entry
        self.redis_server.hset(
            self.user_id_hash_prefix + user_id, mapping=mapping
        )
        lock.release()

        return True

    def update(
        self,
        user_id,
        user_group=None,
        password_hash=None,
        user_role=None,
        permissions=None,
    ):
        """Update the user credentials.

        Renaming an entry is not allowed, only existing entries with the same
        user_id can be updated.

        If a parameter is not provided (set None), the original value will be
        kept.

        HSET User-id-hash db -> set user with all required entries

        Args:
            user_id (str): The user id
            user_group (str): The user group
            password_hash (str): The password hash
            user_role (str): Get the role of the user ("admin", "user",
                             "guest")
            permissions (dict): A dictionary of permissions

        Returns:
            bool:
            True is success, False if user is not in the database

        """
        if (
            self.redis_server.exists(self.user_id_hash_prefix + user_id)
            is False
        ):
            return False

        user_creds = self.get_credentials(user_id)

        if password_hash is None:
            password_hash = user_creds["password_hash"]

        if user_role is None:
            user_role = user_creds["user_role"]

        if user_group is None:
            user_group = user_creds["user_group"]

        if permissions is None:
            permissions = user_creds["permissions"]

        pstring = pickle.dumps(permissions)

        lock = self.redis_server.lock(name="update_user_lock", timeout=1)
        lock.acquire()

        mapping = {
            "user_id": user_id,
            "password_hash": password_hash,
            "user_role": user_role,
            "user_group": user_group,
            "permissions": pstring,
        }
        # Update the database entry
        self.redis_server.hset(
            self.user_id_hash_prefix + user_id, mapping=mapping
        )

        lock.release()

        return True

    def exists(self, user_id):
        """Check if the user is in the database

        Args:
            user_id (str): The user id

        Returns:
            bool:
            True is user exists, False otherwise
        """
        return self.redis_server.exists(self.user_id_hash_prefix + user_id)

    def delete(self, user_id):
        """Remove a user id from the database

        HDEL User-id db
        DEL User-id-hash db

        Args:
            user_id (str): The user id

        Returns:
            bool:
            True is user exists, False otherwise
        """
        if self.exists(user_id) is False:
            return False

        lock = self.redis_server.lock(name="delete_user_lock", timeout=1)
        lock.acquire()
        # Delete the entry from the user id database
        self.redis_server.hdel(self.user_id_db, user_id)
        # Delete the actual user entry
        self.redis_server.delete(self.user_id_hash_prefix + user_id)
        lock.release()

        return True

    def list_all_ids(self):
        """
        List all user id's that are in the database

        HKEYS on the user id database

        Returns:
            list:
            A list of all user ids in the database
        """
        values = []
        l_entries = self.redis_server.hkeys(self.user_id_db)
        print(l_entries)
        for entry in l_entries:
            if entry:
                entry = entry.decode()
            values.append(entry)

        return values


# Create the Redis interface instance
redis_user_interface = RedisUserInterface()


def test_management(r):
    user_id = "Soeren"
    user_group = "test_1"
    password_hash = "hash"
    user_role = "admin"
    permissions = {
        "locations": {
            "NC": {"mapsets": ["PERMANWENT", "user1"]},
            "ECAD": {"mapsets": ["Temp", "Prec"]},
        },
        "modules": ["r.series", "r.slope.aspect"],
    }

    r.delete(user_id)

    r.add(
        user_id=user_id,
        user_group=user_group,
        password_hash=password_hash,
        user_role=user_role,
        permissions=permissions,
    )

    user_creds = r.get_credentials(user_id)
    # print(user_creds)

    if user_creds["user_id"] != user_id:
        raise Exception("add does not work")
    if user_creds["user_group"] != user_group:
        raise Exception("add does not work")
    if user_creds["password_hash"] != password_hash:
        raise Exception("add does not work")
    if user_creds["user_role"] != user_role:
        raise Exception("add does not work")

    if r.get_password_hash(user_id) != password_hash:
        raise Exception("get_password_hash does not work")

    if r.get_role(user_id) != user_role:
        raise Exception("get_role does not work")

    r.update(
        user_id=user_id,
        user_group=user_group,
        password_hash="hello",
        user_role=None,
        permissions=None,
    )

    user_creds = r.get_credentials(user_id)
    # print(user_creds)

    if user_creds["user_id"] != user_id:
        raise Exception("update does not work")
    if user_creds["user_group"] != user_group:
        raise Exception("add does not work")
    if user_creds["password_hash"] != "hello":
        raise Exception("update does not work")
    if user_creds["user_role"] != user_role:
        raise Exception("update does not work")

    user_group = "test_2"
    r.update(
        user_id=user_id,
        user_group=user_group,
        password_hash="yellow",
        user_role="user",
        permissions={
            "locations": {"utm32n": {"mapsets": ["PERMANWENT"]}},
            "modules": [
                "i.vi",
            ],
        },
    )

    user_creds = r.get_credentials(user_id)
    # print(user_creds)

    if user_creds["user_id"] != user_id:
        raise Exception("update does not work")
    if user_creds["user_group"] != user_group:
        raise Exception("add does not work")
    if user_creds["password_hash"] != "yellow":
        raise Exception("update does not work")
    if user_creds["user_role"] != "user":
        raise Exception("update does not work")
    if "utm32n" not in user_creds["permissions"]["locations"]:
        raise Exception("update does not work")

    user_ids = r.list_all_ids()
    if user_id not in user_ids:
        raise Exception("list_all_ids does not work")

    user_id_1 = user_id
    user_id = "Thomas"
    user_id_2 = user_id
    user_role = "guest"

    r.add(
        user_id=user_id,
        user_group=user_group,
        password_hash=password_hash,
        user_role=user_role,
        permissions=permissions,
    )

    user_creds = r.get_credentials(user_id)

    if user_creds["user_id"] != user_id:
        raise Exception("add does not work")
    if user_creds["user_group"] != user_group:
        raise Exception("add does not work")
    if user_creds["password_hash"] != password_hash:
        raise Exception("add does not work")
    if user_creds["user_role"] != user_role:
        raise Exception("user_role does not work")

    if r.get_password_hash(user_id) != password_hash:
        raise Exception("get_password_hash does not work")

    if r.get_role(user_id) != user_role:
        raise Exception("get_role does not work")

    user_ids = r.list_all_ids()
    if user_id not in user_ids:
        raise Exception("list_all_ids does not work")

    for id_ in user_ids:
        print(id_, r.get_credentials(id_))

    r.delete(user_id_1)
    r.delete(user_id_2)

    user_ids = r.list_all_ids()

    if user_ids:
        raise Exception("delete does not work")


if __name__ == "__main__":
    import os
    import signal
    import time

    pid = os.spawnl(
        os.P_NOWAIT, "/usr/bin/redis-server", "./redis.conf", "--port 7000"
    )

    time.sleep(1)

    try:
        r = RedisUserInterface()
        r.connect(host="localhost", port=7000)
        test_management(r)
        r.disconnect()
    except Exception:
        raise
    finally:
        os.kill(pid, signal.SIGTERM)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Storage base class
"""
import os
from google.cloud import bigquery
from google.cloud import storage
import xml.etree.ElementTree as eTree
from .exceptions import GoogleCloudAPIError
import dateutil.parser as dtparser


__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"

GML_BODY = """<?xml version="1.0" encoding="utf-8" ?>
<ogr:FeatureCollection
     xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
     xmlns:ogr="http://ogr.maptools.org/"
     xmlns:gml="http://www.opengis.net/gml">
  <gml:featureMember>
    <ogr:footprint fid="footprint">
      <ogr:geometryProperty>
        <gml:Polygon srsName="EPSG:4326">
          <gml:outerBoundaryIs>
            <gml:LinearRing>
              <gml:coordinates>%s</gml:coordinates>
            </gml:LinearRing>
          </gml:outerBoundaryIs>
        </gml:Polygon>
      </ogr:geometryProperty>
    </ogr:footprint>
  </gml:featureMember>
</ogr:FeatureCollection>
"""


def extract_sensor_id_from_scene_id(scene_id):
    """Extract the sensor id from a Landsat scene id

    Args:
        scene_id (str): The landsat scene id

    Returns:
        (str)
        The sencor id

    """
    return "%s0%s" % (scene_id[0:2], scene_id[2:3])


def get_landsat_query(scene_id, spacecraft_id):
    """Extract the sensor id from a Landsat scene id

    Args:
        scene_id (str): The landsat scene id
        spacecraft_id (str): The landsat spacecraft id

    Returns:
        query (str): The landsat query
        scene_id_query (str): The scene id query
        spacecraft_id_query (str): the spacecraft id query
        has_where_statement (bool): If True scene_id and/or spacecraft_id is
                                    set and a WHERE part in the query is
                                    required

    """
    has_where_statement = False
    query = (
        "SELECT scene_id,sensing_time,north_lat,south_lat,east_lon,"
        "west_lon,cloud_cover,total_size FROM `bigquery-public-data"
        ".cloud_storage_geo_index.landsat_index` "
    )

    if scene_id:
        scene_id_query = "scene_id = '%s'" % scene_id
        has_where_statement = True

    if spacecraft_id:
        spacecraft_id_query = "spacecraft_id = '%s'" % spacecraft_id
        has_where_statement = True
    return query, scene_id_query, spacecraft_id_query, has_where_statement


def get_sentinel_query(scene_id):
    """Extract the sensor id from a Sentinel scene id

    Args:
        scene_id (str): The Sentinel scene id

    Returns:
        query (str): The Sentinel query
        scene_id_query (str): The scene id query
        has_where_statement (bool): If True scene_id is set and a WHERE part
                                    in the query is required

    """
    # Select specific columns from the sentinel table
    query = (
        "SELECT product_id,sensing_time,north_lat,south_lat,east_lon,"
        "west_lon,cloud_cover,total_size FROM `bigquery-public-data"
        ".cloud_storage_geo_index.sentinel_2_index` "
    )

    if scene_id:
        scene_id_query = "product_id = '%s'" % scene_id
        has_where_statement = True
    return query, scene_id_query, has_where_statement


def get_where_query(
    scene_id_query,
    spacecraft_id_query,
    temporal_query,
    spatial_query,
    cloud_cover_query,
):
    """Create where query to request the satellite archive

    Args:
        scene_id_query (str): The scene id query
        spacecraft_id_query (str): The spacecraft id query
        temporal_query (str): The temporal query
        spatial_query (str): The spatial query
        cloud_cover_query (str): The cloud cover query

    Returns:
        query (str): The WHERE part of the query
    """

    query = " WHERE "
    statement_count = 0

    if scene_id_query:
        query += f" ({scene_id_query}) "
        statement_count += 1

    if spacecraft_id_query:
        query += (
            f" AND ({spacecraft_id_query}) "
            if statement_count > 0
            else f" ({spacecraft_id_query}) "
        )
        statement_count += 1

    if temporal_query:
        query += (
            f" AND ({temporal_query}) "
            if statement_count > 0
            else f" ({temporal_query}) "
        )
        statement_count += 1

    if spatial_query:
        query += (
            f" AND ({spatial_query}) "
            if statement_count > 0
            else f" ({spatial_query}) "
        )
        statement_count += 1

    if cloud_cover_query:
        query += (
            f" AND ({cloud_cover_query}) "
            if statement_count > 0
            else f" ({cloud_cover_query}) "
        )
        statement_count += 1
    return query


class GoogleSatelliteBigQueryInterface(object):
    """Query interface to the Sentinel2 and Landsat public geo-data
    available in the google cloud storage
    """

    def __init__(self, config):
        """
        Interface to the Sentinel2 and Landsat data in the google cloud storage

        Args:
            config: The configuration of Actinia Core
        """
        self.aws_sentinel_base_url = "http://sentinel-s2-l1c.s3.amazonaws.com"
        self.gcs_url = "https://storage.googleapis.com/"
        self.sentinel_xml_metadata_file = "MTD_MSIL1C.xml"
        self.config = config
        os.environ[
            "GOOGLE_APPLICATION_CREDENTIALS"
        ] = self.config.GOOGLE_APPLICATION_CREDENTIALS
        os.environ["GOOGLE_CLOUD_PROJECT"] = self.config.GOOGLE_CLOUD_PROJECT

        self.sentinel_bands = [
            "B01",
            "B02",
            "B03",
            "B04",
            "B05",
            "B06",
            "B07",
            "B08",
            "B8A",
            "B09",
            "B10",
            "B11",
            "B12",
        ]

        self.landsat_scene_bands = {
            "LT04": ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "MTL"],
            "LT05": ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "MTL"],
            "LE07": [
                "B1",
                "B2",
                "B3",
                "B4",
                "B5",
                "B6_VCID_2",
                "B6_VCID_1",
                "B7",
                "B8",
                "MTL",
            ],
            "LC08": [
                "B1",
                "B2",
                "B3",
                "B4",
                "B5",
                "B6",
                "B7",
                "B8",
                "B9",
                "B10",
                "B11",
                "MTL",
            ],
        }

        self.raster_suffixes = {
            "LT04": [".1", ".2", ".3", ".4", ".5", ".6", ".7"],
            "LT05": [".1", ".2", ".3", ".4", ".5", ".6", ".7"],
            "LE07": [".1", ".2", ".3", ".4", ".5", ".61", ".62", ".7", ".8"],
            "LC08": [
                ".1",
                ".2",
                ".3",
                ".4",
                ".5",
                ".6",
                ".7",
                ".8",
                ".9",
                ".10",
                ".11",
            ],
        }

    def _start_clients(self):
        self.bigquery_client = bigquery.Client()
        self.storage_client = storage.Client()

    def query_landsat_archive(
        self,
        start_time,
        end_time,
        lat=None,
        lon=None,
        cloud_cover=None,
        scene_id=None,
        spacecraft_id=None,
    ):
        return self._query_satellite_archive(
            satellite="landsat",
            start_time=start_time,
            end_time=end_time,
            lat=lat,
            lon=lon,
            cloud_cover=cloud_cover,
            scene_id=scene_id,
            spacecraft_id=spacecraft_id,
        )

    def query_sentinel2_archive(
        self,
        start_time,
        end_time,
        lat=None,
        lon=None,
        cloud_cover=None,
        scene_id=None,
    ):
        return self._query_satellite_archive(
            satellite="sentinel2",
            start_time=start_time,
            end_time=end_time,
            lat=lat,
            lon=lon,
            cloud_cover=cloud_cover,
            scene_id=scene_id,
        )

    def _query_satellite_archive(
        self,
        satellite,
        start_time,
        end_time,
        lat=None,
        lon=None,
        cloud_cover=None,
        scene_id=None,
        spacecraft_id=None,
    ):
        """
        Query the landsat or sentinel archive by time interval and coordinates
        and return a list of scenes with metadata

        Args:
            satellite (str): The satellite that should be used for querying
                             landsat or sentinel2
            start_time (datetime.datetime): datetime object of the start time
                                            to query
            end_time (datetime.datetime): datetime object of the end time to
                                          query
            lat (float): Latitude coordinates that should intersect the landsat
                         scenes
            lon (float): Longitude coordinates that should intersect the
                         landsat scenes
            cloud_cover (float): The maximum cloud cover 0-100
            scene_id (str): The scene id to search for
            spacecraft_id (str): The spacecraft id of the landsat scene

        Returns:
            (list)
            A list of dict
            [
              {
                "cloud_cover": 99.0,
                "east_lon": 140.41507,
                "north_lat": -6.28158,
                "scene_id": "LE71010652001001EDC01",
                "sensing_time": "2001-01-01T00:41:06.1272135Z",
                "south_lat": -8.17417,
                "total_size": 74683859,
                "west_lon": 138.29205
              },
              {
                "cloud_cover": 73.0,
                "east_lon": 156.75294,
                "north_lat": 52.69982,
                "scene_id": "LE71010242001001EDC01",
                "sensing_time": "2001-01-01T00:24:46.2473502Z",
                "south_lat": 50.67277,
                "total_size": 161222513,
                "west_lon": 153.20544
              }
            ]
        """
        try:
            self._start_clients()
            scene_id_query = None
            spacecraft_id_query = None
            temporal_query = None
            spatial_query = None
            cloud_cover_query = None
            has_where_statement = False

            if satellite == "landsat":
                (
                    query,
                    scene_id_query,
                    spacecraft_id_query,
                    has_where_statement,
                ) = get_landsat_query(scene_id, spacecraft_id)
            else:
                (
                    query,
                    scene_id_query,
                    has_where_statement,
                ) = get_sentinel_query(scene_id)

            if start_time and end_time:
                start_time = dtparser.parse(start_time)
                end_time = dtparser.parse(end_time)
                temporal_query = (
                    "sensing_time >= '%(start)s' "
                    "AND sensing_time <= '%(end)s'"
                    % {
                        "start": start_time.isoformat(),
                        "end": end_time.isoformat(),
                    }
                )
                has_where_statement = True

            if lon and lat:
                spatial_query = (
                    "west_lon <= %(lon)f AND east_lon >= %(lon)f AND "
                    "north_lat >= %(lat)f AND south_lat <= %(lat)f"
                    % {"lon": float(lon), "lat": float(lat)}
                )
                has_where_statement = True

            if cloud_cover:
                cloud_cover_query = "cloud_cover <= %s" % str(cloud_cover)
                has_where_statement = True

            if has_where_statement is True:
                query += get_where_query(
                    scene_id_query,
                    spacecraft_id_query,
                    temporal_query,
                    spatial_query,
                    cloud_cover_query,
                )

            query_results = self.bigquery_client.query(query)
            result = []

            if query_results.result():
                for row in query_results.result():
                    (
                        scene_id,
                        sensing_time,
                        north_lat,
                        south_lat,
                        east_lon,
                        west_lon,
                        cloud_cover,
                        total_size,
                    ) = row
                    result.append(
                        dict(
                            scene_id=scene_id,
                            sensing_time=sensing_time,
                            north_lat=north_lat,
                            south_lat=south_lat,
                            east_lon=east_lon,
                            west_lon=west_lon,
                            cloud_cover=cloud_cover,
                            total_size=total_size,
                        )
                    )

            return result

        except Exception as e:
            raise GoogleCloudAPIError(
                "An error occurred while querying "
                "google archive. Error message: %s" % str(e)
            )

    def get_landsat_urls(self, scene_ids, bands=None):
        """
        Receive the Google Cloud Storage (GCS) download urls and time stamps
        for a list of Landsat scene ids with additional metadata.

        The download urls include the public address and the google cloud
        storage address. The resulting dictionary has as keys the scene ids
        that contain the file name, the tile name and the urls per band.

        Args:
            scene_ids: A list of scene ids that should be collected
            bands: The bands to create the download urls for

        Returns:
            (dict)
            A dictionary of scene entries

            Example:

            {u'LC80440342013106LGN01': {
                'B1': {'map': u'LC80440342013106LGN01_B1',
                        'gcs_url': u'gs://gcp-public-data-landsat/LC08/PRE/044'
                                    '/034/LC80440342013106LGN01/LC804403420131'
                                    '06LGN01_B1.TIF',
                        'public_url': u'https://storage.googleapis.com/gcp-'
                                       'public-data-landsat/LC08/PRE/044/034/'
                                       'LC80440342013106LGN01/LC80440342013106'
                                       'LGN01_B1.TIF',
                        'file': u'LC80440342013106LGN01_B1.TIF'},
                'MTL': {'map': u'LC80440342013106LGN01_MTL',
                        'gcs_url': u'gs://gcp-public-data-landsat/LC08/PRE/044'
                                    '/034/LC80440342013106LGN01/LC804403420131'
                                    '06LGN01_MTL.txt',
                        'public_url': u'https://storage.googleapis.com/gcp-'
                                       'public-data-landsat/LC08/PRE/044/034/'
                                       'LC80440342013106LGN01/LC80440342013106'
                                       'LGN01_MTL.txt',
                        'file': u'LC80440342013106LGN01_MTL.txt'},
                'timestamp': u'2013-04-16T18:47:53.5577434Z'},
             u'LC80440342016259LGN00': {
                'B1': {'map': u'LC80440342016259LGN00_B1',
                       'gcs_url': u'gs://gcp-public-data-landsat/LC08/PRE/044/'
                                   '034/LC80440342016259LGN00/LC80440342016259'
                                   'LGN00_B1.TIF',
                       'public_url': u'https://storage.googleapis.com/gcp-'
                                      'public-data-landsat/LC08/PRE/044/034/'
                                      'LC80440342016259LGN00/LC80440342016259'
                                      'LGN00_B1.TIF',
                       'file': u'LC80440342016259LGN00_B1.TIF'},
                'MTL': {'map': u'LC80440342016259LGN00_MTL',
                        'gcs_url': u'gs://gcp-public-data-landsat/LC08/PRE/044'
                                    '/034/LC80440342016259LGN00/LC804403420162'
                                    '59LGN00_MTL.txt',
                        'public_url': u'https://storage.googleapis.com/gcp-'
                                       'public-data-landsat/LC08/PRE/044/034/'
                                       'LC80440342016259LGN00/'
                                       'LC80440342016259LGN00_MTL.txt',
                        'file': u'LC80440342016259LGN00_MTL.txt'},
                'timestamp': u'2016-09-15T18:46:18.6867380Z'}}

        """

        # Google BigQuery landsat table SQL query
        # SELECT * FROM [bigquery-public-data:cloud_storage_geo_index.
        # landsat_index]
        #          WHERE sensing_time > '2000-01-01T00:00:00' LIMIT 1;
        #
        # [
        #   {
        #     "scene_id": "LE72210652016307CUB00",
        #     "product_id": "LE07_L1TP_221065_20161102_20161128_01_T1",
        #     "spacecraft_id": "LANDSAT_7",
        #     "sensor_id": "ETM",
        #     "date_acquired": "2016-11-02",
        #     "sensing_time": "2016-11-02T13:14:58.9749958Z",
        #     "collection_number": "01",
        #     "collection_category": "T1",
        #     "data_type": "L1TP",
        #     "wrs_path": "221",
        #     "wrs_row": "65",
        #     "cloud_cover": "1.0",
        #     "north_lat": "-6.28415",
        #     "south_lat": "-8.16646",
        #     "west_lon": "-47.15011",
        #     "east_lon": "-44.96096",
        #     "total_size": "235109281",
        #     "base_url": "gs://gcp-public-data-landsat/LE07/01/221/065/"
        #                 "LE07_L1TP_221065_20161102_20161128_01_T1"
        #   }
        # ]

        # Assign default bands
        if bands is None:
            bands = ["B1", "B2"]

        try:
            self._start_clients()

            for scene_id in scene_ids:
                sensor_id = extract_sensor_id_from_scene_id(scene_id)
                for band in bands:
                    if band not in self.landsat_scene_bands[sensor_id]:
                        raise Exception(
                            "Unknown landsat band name <%s>" % band
                        )

            # Select specific columns from the sentinel table
            query = (
                "SELECT scene_id,sensing_time,base_url FROM "
                "`bigquery-public-data.cloud_storage_geo_index.landsat_index` "
                'WHERE scene_id IN ("%s");' % '","'.join(scene_ids)
            )

            query_results = self.bigquery_client.query(query)
            result = {}

            if query_results.result():
                for row in query_results.result():
                    scene_id, sensing_time, base_url = row
                    public_url = self.gcs_url + base_url[5:]
                    sensor_id = extract_sensor_id_from_scene_id(scene_id)

                    result[scene_id] = {}
                    result[scene_id]["timestamp"] = sensing_time

                    for band in bands:
                        suffix = "TIF"
                        if band == "MTL":
                            suffix = "txt"

                        file_name = "%s_%s.%s" % (scene_id, band, suffix)
                        map_name = "%s_%s" % (scene_id, band)
                        if band != "MTL":
                            index = self.landsat_scene_bands[sensor_id].index(
                                band
                            )
                            raster_suffix = self.raster_suffixes[sensor_id][
                                index
                            ]
                            map_name = "%s%s" % (scene_id, raster_suffix)

                        result[scene_id][band] = {}
                        result[scene_id][band]["file"] = file_name
                        result[scene_id][band]["map"] = map_name
                        result[scene_id][band]["public_url"] = (
                            public_url + "/" + file_name
                        )
                        result[scene_id][band]["gcs_url"] = (
                            base_url + "/" + file_name
                        )

            return result

        except Exception as e:
            raise GoogleCloudAPIError(
                "An error occurred while fetching "
                "Landsat download URL's. Error message: %s" % str(e)
            )

    def get_sentinel_urls(self, product_ids, bands=None):
        """Receive the download urls and time stamps for a list of Sentinel2
        product ids from Google Big Query service

        The download urls include the public address and the google cloud
        storage address. The resulting dictionary has as keys the product ids
        that contain the file name, the tile name and the urls per band. In
        addition the time stamp, the GML footprint and the XML download path
        is included.

        Args:
            product_ids: A list of sentinel product ids
            bands: A list of band names

        Returns:
            dict
            A dictionary that contains the time stamp, the tile names, the map
            names in GRASS the goggle cloud storage urls, the public urls, the
            GML footprint and the XML file download path.

            Boundingbox = (min_x, max_y, max_x, min_y)

            Example:

            u'S2A_MSIL1C_20170301T225331_N0204_R115_T56DNG_20170301T225347': {
                'B04': {'file': u'S2A_MSIL1C_20170301T225331_N0204_R115_T56DNG'
                                 '_20170301T225347_B04',
                        'gcs_url': u'gs://gcp-public-data-sentinel-2/tiles/56/'
                                    'D/NG/S2A_MSIL1C_20170301T225331_N0204_'
                                    'R115_T56DNG_20170301T225347.SAFE/GRANULE/'
                                    'L1C_T56DNG_A008835_20170301T225347/'
                                    'IMG_DATA/T56DNG_20170301T225331_B04.jp2',
                        'public_url': u'https://storage.googleapis.com/gcp-'
                                       'public-data-sentinel-2/tiles/56/D/NG/'
                                       'S2A_MSIL1C_20170301T225331_N0204_R115_'
                                       'T56DNG_20170301T225347.SAFE/GRANULE/'
                                       'L1C_T56DNG_A008835_20170301T225347/'
                                       'IMG_DATA/T56DNG_20170301T225331_'
                                       'B04.jp2',
                        'tile': u'T56DNG_20170301T225331_B04.jp2'},
                'B08': {'file': u'S2A_MSIL1C_20170301T225331_N0204_R115_T56DNG'
                                 '_20170301T225347_B08',
                        'gcs_url': u'gs://gcp-public-data-sentinel-2/tiles/56/'
                                    'D/NG/S2A_MSIL1C_20170301T225331_N0204_'
                                    'R115_T56DNG_20170301T225347.SAFE/GRANULE/'
                                    'L1C_T56DNG_A008835_20170301T225347/IMG_'
                                    'DATA/T56DNG_20170301T225331_B08.jp2',
                        'public_url': u'https://storage.googleapis.com/gcp-'
                                       'public-data-sentinel-2/tiles/56/D/NG/'
                                       'S2A_MSIL1C_20170301T225331_N0204_R115_'
                                       'T56DNG_20170301T225347.SAFE/GRANULE/'
                                       'L1C_T56DNG_A008835_20170301T225347/'
                                       'IMG_DATA/T56DNG_20170301T225331_'
                                       'B08.jp2',
                        'tile': u'T56DNG_20170301T225331_B08.jp2'},
                'bbox': (-70.45384256042797,
                         153.19042365887816,
                         -70.30447316127372,
                         152.99946420488922),
                'gcs_xml_metadata_url': u'gs://gcp-public-data-sentinel-2/'
                                         'tiles/56/D/NG/S2A_MSIL1C_20170301T22'
                                         '5331_N0204_R115_T56DNG_20170301T2253'
                                         '47.SAFE/MTD_MSIL1C.xml',
                'gml_footprint': '<?xml version="1.0" encoding="utf-8" ?>\n'
                                 '<ogr:FeatureCollection\n'
                                 '     xmlns:xsi="http://www.w3.org/2001/'
                                 'XMLSchema-instance"\n'
                                 '     xmlns:ogr="http://ogr.maptools.org/"\n'
                                 '     xmlns:gml="http://www.opengis.net/gml">'
                                 '\n  <gml:featureMember>\n'
                                 '    <ogr:footprint fid="footprint">\n'
                                 '      <ogr:geometryProperty>'
                                 '<gml:Polygon srsName="EPSG:4326">'
                                 '<gml:outerBoundaryIs>'
                                 '<gml:LinearRing>'
                                 '<gml:coordinates>153.19042365887816,'
                                 '-70.30447316127372 153.08861867314772,'
                                 '-70.38450125221787 152.99946420488922,'
                                 '-70.45384256042797 152.99946824095414,'
                                 '-70.30601749291662 153.19042365887816,'
                                 '-70.30447316127372 153.19042365887816,'
                                 '-70.30447316127372</gml:coordinates>'
                                 '</gml:LinearRing></gml:outerBoundaryIs>'
                                 '</gml:Polygon></ogr:geometryProperty>\n'
                                 '    </ogr:footprint>\n'
                                 '  </gml:featureMember>\n'
                                 '</ogr:FeatureCollection>\n',
                'public_xml_metadata_url': u'https://storage.googleapis.com/'
                                            'gcp-public-data-sentinel-2/tiles/'
                                            '56/D/NG/S2A_MSIL1C_20170301T2253'
                                            '31_N0204_R115_T56DNG_20170301'
                                            'T225347.SAFE/MTD_MSIL1C.xml',
                'timestamp': u'2017-03-01T22:53:47.503000Z'}

        """

        # Google BigQuery landsat table SQL query
        # SELECT * FROM [bigquery-public-data:cloud_storage_geo_index.
        # sentinel_2_index]
        #  where product_id =
        #  "S2A_MSIL1C_20170208T092131_N0204_R093_T35TLF_20170208T092143"
        #
        # Result:
        # [
        #   {
        #     "granule_id": "L1C_T35TLF_A008527_20170208T092143",
        #     "product_id": "S2A_MSIL1C_20170208T092131_N0204_R093_T35TLF_"
        #                   "20170208T092143",
        #     "datatake_identifier": "GS2A_20170208T092131_008527_N02.04",
        #     "mgrs_tile": "35TLF",
        #     "sensing_time": "2017-02-08T09:21:43.890000Z",
        #     "geometric_quality_flag": "PASSED",
        #     "generation_time": "2017-02-08T09:21:43.000000Z",
        #     "north_lat": "41.5297704546",
        #     "south_lat": "41.1784693962",
        #     "west_lon": "24.6026969474",
        #     "east_lon": "24.7426671934",
        #     "base_url": "gs://gcp-public-data-sentinel-2/tiles/35/T/LF/S2A"
        #                 "_MSIL1C_20170208T092131_N0204_R093_T35TLF_"#
        #                 "20170208T092143.SAFE",
        #     "total_size": "30544232",
        #     "cloud_cover": "0.0"
        #   }
        # ]
        # Public link:
        # https://storage.googleapis.com/gcp-public-data-sentinel-2/tiles/35/
        # T/LF/S2A_MSIL1C_20170208T092131_N0204_R093_T35TLF_20170208T092143.
        # SAFE/GRANULE/L1C_T35TLF_A008527_20170208T092143/IMG_DATA/T35TLF_
        # 20170208T092131_B01.jp2

        # Assign default bands
        if bands is None:
            bands = ["B04", "B08"]

        try:
            self._start_clients()

            for band in bands:
                if band not in self.sentinel_bands:
                    raise Exception("Unknown Sentinel-2 band name <%s>" % band)

            # Select specific columns from the sentinel table
            query = (
                "SELECT granule_id,product_id,sensing_Time,datatake_identifier"
                ",base_url FROM `bigquery-public-data.cloud_storage_geo_index"
                ".sentinel_2_index` "
                'WHERE product_id IN ("%s");' % '","'.join(product_ids)
            )

            rows = list(self.bigquery_client.query(query))  # API request

            result = {}

            if rows:
                for row in rows:
                    (
                        granule_id,
                        product_id,
                        sensing_time,
                        datatake_identifier,
                        base_url,
                    ) = row
                    tile_base_name = (
                        granule_id[4:10] + datatake_identifier[4:20]
                    )
                    gcs_url = (
                        base_url + "/GRANULE/" + granule_id + "/IMG_DATA/"
                    )
                    public_url = (
                        self.gcs_url
                        + base_url[5:]
                        + "/GRANULE/"
                        + granule_id
                        + "/IMG_DATA/"
                    )

                    result[product_id] = {}
                    result[product_id]["timestamp"] = sensing_time
                    result[product_id]["public_xml_metadata_url"] = (
                        self.gcs_url
                        + base_url[5:]
                        + "/"
                        + self.sentinel_xml_metadata_file
                    )
                    result[product_id]["gcs_xml_metadata_url"] = (
                        base_url + "/" + self.sentinel_xml_metadata_file
                    )

                    # Generate the GML file from the sentinel product footprint
                    # The whole XML content is returned as well
                    (
                        gml,
                        xml_metadata,
                        bbox,
                    ) = self._generate_sentinel2_footprint(base_url=base_url)
                    result[product_id]["gml_footprint"] = gml
                    result[product_id]["bbox"] = bbox
                    # The xml content is currently not needed
                    # result[product_id]["xml_metadata"] = xml_metadata

                    for band in bands:
                        tile_name = "%s_%s.jp2" % (tile_base_name, band)
                        map_name = "%s_%s" % (product_id, band)

                        result[product_id][band] = {}
                        result[product_id][band]["tile"] = tile_name
                        result[product_id][band]["file"] = map_name
                        result[product_id][band]["public_url"] = (
                            public_url + tile_name
                        )
                        result[product_id][band]["gcs_url"] = (
                            gcs_url + tile_name
                        )

            return result

        except Exception as e:
            raise GoogleCloudAPIError(
                "An error occurred while fetching "
                "Sentinel-2 download URL's. Error message: %s" % str(e)
            )

    def _generate_sentinel2_footprint(self, base_url):
        """Download the sentinel XML metadata and parse it for the footprint

        Args:
            base_url: The google cloud storage base url of the required
                      product_id

        Returns: a tuple of strings
            (str, str)
            The first string is the footprint as GML code,
            The second string the the metadata XML document

        """

        # Download the XML file from the google cloud storage using the cloud
        # API
        bucket = self.storage_client.get_bucket("gcp-public-data-sentinel-2")
        blob = bucket.blob(
            base_url.replace("gs://gcp-public-data-sentinel-2/", "")
            + "/"
            + self.sentinel_xml_metadata_file
        )

        xml_content = blob.download_as_string()

        # Find the coordinates in the XML string
        root = eTree.fromstring(xml_content)
        # The namespace will hopefully not change
        geo_info = root.find(
            "{https://psd-14.sentinel2.eo.esa.int/PSD/"
            "User_Product_Level-1C.xsd}Geometric_Info"
        )
        global_footprint = (
            geo_info.find("Product_Footprint")
            .find("Product_Footprint")
            .find("Global_Footprint")
        )
        coordinates = global_footprint.find("EXT_POS_LIST").text

        # Extract the coordinates from the text and convert it into lat/lon
        # tuples
        coord_list = coordinates.split()
        gml_coord_list = []

        min_x = 10000000
        max_x = -10000000
        min_y = 10000000
        max_y = -10000000

        i = 0
        while i < len(coord_list):
            y = coord_list[i]
            if float(y) > max_y:
                max_y = float(y)
            if float(y) < min_y:
                min_y = float(y)
            i += 1
            x = coord_list[i]
            if float(x) > max_x:
                max_x = float(x)
            if float(x) < min_x:
                min_x = float(x)
            i += 1
            gml_coord_list.append("%s,%s" % (x, y))

        return (
            GML_BODY % " ".join(gml_coord_list),
            xml_content,
            (min_x, max_y, max_x, min_y),
        )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster colors management

TODO: Maybe more tests required, test_raster_colors.py is in place and works
"""

from tempfile import NamedTemporaryFile
from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralRasterColorsOutput(EphemeralProcessing):
    def __init__(self, *args):
        EphemeralProcessing.__init__(self, *args)

    def _execute(self, skip_permission_check=True):
        self._setup()

        raster_name = self.map_name
        self.required_mapsets.append(self.mapset_name)

        with NamedTemporaryFile(
            mode="w+", suffix=".color", dir=self.temp_file_path
        ) as file:
            result_file = file.name

        self.request_data = {
            "1": {"module": "r.colors.out", "inputs": {"map": "", "rules": ""}}
        }
        self.request_data["1"]["inputs"]["map"] = (
            raster_name + "@" + self.mapset_name
        )
        self.request_data["1"]["inputs"]["rules"] = result_file

        # Run the selected modules
        process_list = (
            self._create_temporary_grass_environment_and_process_list(
                skip_permission_check=True
            )
        )
        self._execute_process_list(process_list)

        with open(result_file, "r") as file:
            result = file.read().strip().split("\n")

        self.module_results = result



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Async process mapset test case admin
"""
import unittest
from flask.json import dumps as json_dumps
from flask.json import loads as json_load

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Guido Riembauer, Anika Weinmann"
__copyright__ = "Copyright 2021, mundialis GmbH & Co. KG"
__maintainer__ = "mundialis"

mapset_name = "test_strds_merging"

process_chain_create_strds1 = {
    "list": [
        {
            "module": "g.copy",
            "id": "g_copy_raster_1",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2015001.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2015001.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "g.copy",
            "id": "g_copy_raster_2",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2015032.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2015032.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "g.copy",
            "id": "g_copy_raster_3",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2015060.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2015060.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "t.create",
            "id": "t.create_strds",
            "inputs": [
                {"param": "output", "value": "modis"},
                {"param": "title", "value": "modis"},
                {"param": "description", "value": "modis"},
            ],
        },
        {
            "module": "t.register",
            "id": "t.register_strds",
            "inputs": [
                {"param": "input", "value": "modis"},
                {
                    "param": "maps",
                    "value": "MOD11B3.A2015001.h11v05.single_LST_Day_6km,"
                    "MOD11B3.A2015032.h11v05.single_LST_Day_6km,MOD11B3."
                    "A2015060.h11v05.single_LST_Day_6km",
                },
                {"param": "start", "value": "2015-01-01"},
                {"param": "increment", "value": "1 months"},
            ],
        },
    ],
    "version": "1",
}
process_chain_create_strds2 = {
    "list": [
        {
            "module": "g.copy",
            "id": "g_copy_raster_1",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2016001.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2016001.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "g.copy",
            "id": "g_copy_raster_2",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2016032.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2016032.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "t.create",
            "id": "t.create_strds_modis",
            "inputs": [
                {"param": "output", "value": "modis2"},
                {"param": "title", "value": "modis2"},
                {"param": "description", "value": "modis2"},
            ],
        },
        {
            "module": "t.register",
            "id": "t.register_strds_modis",
            "inputs": [
                {"param": "input", "value": "modis2"},
                {
                    "param": "maps",
                    "value": "MOD11B3.A2016001.h11v05.single_LST_Day_6km,"
                    "MOD11B3.A2016032.h11v05.single_LST_Day_6km",
                },
                {"param": "start", "value": "2016-01-01"},
                {"param": "increment", "value": "1 months"},
            ],
        },
    ],
    "version": "1",
}
process_chain_create_strds3 = {
    "list": [
        {
            "module": "g.copy",
            "id": "g_copy_raster_1",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2016001.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2016001.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "g.copy",
            "id": "g_copy_raster_2",
            "inputs": [
                {
                    "param": "raster",
                    "value": "MOD11B3.A2016032.h11v05.single_LST_Day_6km@modis"
                    "_lst,MOD11B3.A2016032.h11v05.single_LST_Day_6km",
                }
            ],
        },
        {
            "module": "t.create",
            "id": "t.create_strds_modis2",
            "inputs": [
                {"param": "output", "value": "modis2"},
                {"param": "title", "value": "modis2"},
                {"param": "description", "value": "modis2"},
            ],
        },
        {
            "module": "t.register",
            "id": "t.register_strds_modis2",
            "inputs": [
                {"param": "input", "value": "modis2"},
                {
                    "param": "maps",
                    "value": "MOD11B3.A2016001.h11v05.single_LST_Day_6km,"
                    "MOD11B3.A2016032.h11v05.single_LST_Day_6km",
                },
                {"param": "start", "value": "2016-01-01"},
                {"param": "increment", "value": "1 months"},
            ],
        },
        {
            "module": "t.create",
            "id": "t.create_strds_modis_changed",
            "inputs": [
                {"param": "output", "value": "modis"},
                {"param": "title", "value": "modis"},
                {"param": "description", "value": "modis"},
            ],
        },
        {
            "module": "t.register",
            "id": "t.register_strds_modis_changed",
            "inputs": [
                {"param": "input", "value": "modis"},
                {
                    "param": "maps",
                    "value": "MOD11B3.A2016001.h11v05.single_LST_Day_6km,"
                    "MOD11B3.A2016032.h11v05.single_LST_Day_6km",
                },
                {"param": "start", "value": "2016-01-01"},
                {"param": "increment", "value": "1 months"},
            ],
        },
    ],
    "version": "1",
}
process_chain_list = {
    "list": [
        {
            "module": "t.list",
            "id": "t_list_1",
            "inputs": [{"param": "type", "value": "strds"}],
        },
        {
            "module": "t.rast.list",
            "id": "t.rast.list_1",
            "inputs": [{"param": "input", "value": f"modis@{mapset_name}"}],
        },
    ],
    "version": "1",
}


class AsyncMapsetMergingSTRDS(ActiniaResourceTestCaseBase):
    user_mapset = mapset_name
    raster_dict_modis = {
        "MOD11B3.A2015001.h11v05.single_LST_Day_6km": "2015-01-01 00:00:00",
        "MOD11B3.A2015032.h11v05.single_LST_Day_6km": "2015-02-01 00:00:00",
        "MOD11B3.A2015060.h11v05.single_LST_Day_6km": "2015-03-01 00:00:00",
    }
    raster_dict_modis2 = {
        "MOD11B3.A2016001.h11v05.single_LST_Day_6km": "2016-01-01 00:00:00",
        "MOD11B3.A2016032.h11v05.single_LST_Day_6km": "2016-02-01 00:00:00",
    }

    def tearDown(self):
        # unlock and delete the test mapsets
        rv = self.server.get(
            URL_PREFIX + "/locations/nc_spm_08/mapsets",
            headers=self.user_auth_header,
        )
        existing_mapsets = json_load(rv.data)["process_results"]
        if self.user_mapset in existing_mapsets:
            rvdellock = self.server.delete(
                URL_PREFIX
                + "/locations/nc_spm_08/mapsets/%s/lock" % self.user_mapset,
                headers=self.admin_auth_header,
            )
            print(rvdellock.data.decode())

            rvdel = self.server.delete(
                URL_PREFIX
                + "/locations/nc_spm_08/mapsets/%s" % self.user_mapset,
                headers=self.admin_auth_header,
            )
            print(rvdel.data.decode())

    def check_strds_in_mapset(self, strds_names):
        # check if strds 'modis' is in mapset
        rv = dict()
        for strds_name in strds_names:
            rv[strds_name] = self.server.get(
                URL_PREFIX
                + f"/locations/nc_spm_08/mapsets/{self.user_mapset}/strds",
                headers=self.user_auth_header,
            )
            strds = json_load(rv[strds_name].data)["process_results"]
            self.assertIn(
                strds_name,
                strds,
                f"STRDS '{strds_name}' is not in test mapset",
            )

    def check_modis_strds(self, raster_dict, strds_name):
        # check if correct maps are listed in strds strds
        rv = self.server.get(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            f"strds/{strds_name}/raster_layers",
            headers=self.user_auth_header,
        )
        strds_rasters = json_load(rv.data)["process_results"]
        raster_times = {
            rast["id"].split("@")[0]: rast["start_time"]
            for rast in strds_rasters
        }
        self.assertEqual(
            len(raster_times),
            len(raster_dict),
            f"Number of raster in STRDS are not {len(raster_dict)}",
        )
        self.assertEqual(
            raster_times,
            raster_times,
            "STRDS times and raster names are not correct in test mapset",
        )

    def test_create_strds_in_persistent_user_db(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds1),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        # check if strds 'modis' is in mapset
        self.check_strds_in_mapset(["modis"])
        # check if correct maps are listed i strds 'modis'
        self.check_modis_strds(self.raster_dict_modis, "modis")

    def test_create_strds_in_persistent_user_db_and_list_it(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds1),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        rv2 = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_list),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv2,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        # check if strds 'modis' is in mapset
        self.check_strds_in_mapset(["modis"])
        # check if correct maps are listed i strds 'modis'
        self.check_modis_strds(self.raster_dict_modis, "modis")

    def test_create_strds_in_persistent_user_db_2(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds1),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds2),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        # check if strds 'modis' and 'modis2' is in mapset
        self.check_strds_in_mapset(["modis", "modis2"])
        # check if correct maps are listed in strds 'modis' and 'modis2'
        self.check_modis_strds(self.raster_dict_modis, "modis")
        self.check_modis_strds(self.raster_dict_modis2, "modis2")

    def test_create_strds_in_persistent_user_db_3(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds1),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        # check if strds 'modis' is in mapset
        self.check_strds_in_mapset(["modis"])
        # check if correct maps are listed in strds 'modis'
        self.check_modis_strds(self.raster_dict_modis, "modis")

        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/{self.user_mapset}/"
            "processing_async",
            headers=self.admin_auth_header,
            data=json_dumps(process_chain_create_strds3),
            content_type="application/json",
        )

        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )

        # check if strds 'modis' and 'modis2' is in mapset
        self.check_strds_in_mapset(["modis", "modis2"])
        # check if correct maps are listed in strds 'modis' and 'modis2'
        self.check_modis_strds(self.raster_dict_modis2, "modis")
        self.check_modis_strds(self.raster_dict_modis2, "modis2")


if __name__ == "__main__":
    unittest.main()



#!flask/bin/python
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Returns the version information and the roles that are activated
"""

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"

from flask import make_response, jsonify, request
import os
import re
import importlib
import subprocess
import sys
from actinia_api import URL_PREFIX

from actinia_core.core.common.app import flask_app
from actinia_core.core.common.config import global_config
from actinia_core.core.logging_interface import log
from actinia_core.models.response_models import LinkResponseModel
from . import __version__


G_VERSION = {}
PLUGIN_VERSIONS = {}
PYTHON_VERSION = ""
API_VERSION = ""


def init_versions():
    global PYTHON_VERSION
    global API_VERSION

    g_version = subprocess.run(
        [
            "grass",
            "--tmp-location",
            "epsg:4326",
            "--exec",
            "g.version",
            "-rge",
        ],
        capture_output=True,
    ).stdout
    log.debug("Detecting GRASS GIS version")
    for i in g_version.decode("utf-8").strip("\n").split("\n"):
        G_VERSION[i.split("=")[0]] = i.split("=")[1]

    log.debug("Detecting Plugin versions")
    for i in global_config.PLUGINS:
        module = importlib.import_module(i)
        PLUGIN_VERSIONS[i] = module.__version__

    log.debug("Detecting API versions")
    module = importlib.import_module("actinia_api")
    API_VERSION = module.__version__

    PYTHON_VERSION = sys.version.replace("\n", "- ")


def valid_additional_version_info_key(cand):
    """Checks whether the input is a valid key for inclusion in the version/
    output. Keys may only contain lowercase letters and the underscore, have
    a length between 1 and 25 and must not be any of the reserved keys which
    we will set and populate."""
    minlength = 1
    maxlength = 25
    valid_reg_ex = "^[a-z_]{" + str(minlength) + "," + str(maxlength) + "}$"
    reserved_keys = [
        "grass_version",
        "plugin_versions",
        "plugins",
        "python_version",
        "version",
        "api_version",
    ]
    return cand and cand not in reserved_keys and re.match(valid_reg_ex, cand)


def parse_additional_version_info(env_value):
    """Parses key:value pairs which are separated with pipes and returns them
    as dictionary."""
    additional_info = {}
    # split at pipe |
    kvp_pairs = env_value.split("|")
    for kvp_pair in kvp_pairs:
        # split at colon : but only max once
        kvp = kvp_pair.split(":", 1)
        if len(kvp) == 2 and valid_additional_version_info_key(kvp[0]):
            additional_info[kvp[0]] = kvp[1]
    return additional_info


def find_additional_version_info():
    """If there is a ACTINIA_ADDITIONAL_VERSION_INFO environment variable,
    we'll try to parse its value and return more information that will
    eventually be added to the version output"""
    env_name = "ACTINIA_ADDITIONAL_VERSION_INFO"
    if env_name in os.environ:
        return parse_additional_version_info(os.environ[env_name])
    return {}


def find_running_since_info():
    """If there is a ACTINIA_RUNNING_SINCE environment variable, return its
    content"""
    env_name = "ACTINIA_RUNNING_SINCE"
    if env_name in os.environ:
        return os.environ[env_name]
    return "n/a"


# Return the version of Actinia Core as REST API call
@flask_app.route(URL_PREFIX + "/version")
def version():
    """Return the version information and the roles that are activated

    Returns: Response

    """
    # start with any potential additional version info or an empty dictionary
    info = find_additional_version_info()
    info["version"] = __version__
    info["plugins"] = ",".join(global_config.PLUGINS)
    info["grass_version"] = G_VERSION
    info["plugin_versions"] = PLUGIN_VERSIONS
    info["api_version"] = API_VERSION
    info["python_version"] = PYTHON_VERSION
    info["running_since"] = find_running_since_info()

    return make_response(jsonify(info), 200)


# Return a hint that this version is outdated
@flask_app.route("/api/v1/<path:actinia_path>")
def hint(actinia_path):
    """Return a hint that this version is no longer installed. If an older
    version is installed, this endpoint will be overwritten by a proxy.

    Returns: Response

    """
    url = request.url_root.strip("/") + URL_PREFIX + "/" + actinia_path

    return make_response(
        jsonify(
            LinkResponseModel(
                status="Not found",
                message=(
                    "Are you looking for the current API version? "
                    "Change 'v1' to 'v3' in the URL."
                ),
                links=[url],
            ),
            404,
        )
    )


@flask_app.route("/api/v2/<path:actinia_path>")
def hint_v2(actinia_path):
    """Return a hint that this version is no longer installed. If an older
    version is installed, this endpoint will be overwritten by a proxy.

    Returns: Response

    """
    url = request.url_root.strip("/") + URL_PREFIX + "/" + actinia_path

    return make_response(
        jsonify(
            LinkResponseModel(
                status="Not found",
                message=(
                    "Are you looking for the current API version? "
                    "Change 'v2' to 'v3' in the URL."
                ),
                links=[url],
            ),
            404,
        )
    )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Running a custom UNIX process outside the GRASS GIS environment
"""
from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)
from actinia_core.core.common.exceptions import AsyncProcessError
from actinia_core.rest.base.user_auth import (
    check_location_mapset_module_access,
)
from actinia_core.core.common.process_object import Process

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralCustomProcessing(EphemeralProcessing):
    """Run a custom process asynchronously"""

    def __init__(self, *args):
        """Setup the variables of this class

        Args:
            rdc (ResourceDataContainer): The data container that contains all
                                         required variables for processing
        """

        EphemeralProcessing.__init__(self, *args)
        self.executable = self.rdc.user_data
        self.executable_params = self.rdc.request_data

        self._update_num_of_steps(1)

    def _execute(self, skip_permission_check=False):
        """Overwrite this function in subclasses

        Setup the user credentials, check the executable access and run the
        process
        """
        # Setup the user credentials and logger
        self._setup(init_grass=False)

        # Check if the user has access to the required process executable
        resp = check_location_mapset_module_access(
            self.user_credentials, self.config, module_name=self.executable
        )

        if resp is not None:
            raise AsyncProcessError(
                "Executable <%s> is not supported" % self.executable
            )

        p = Process(
            exec_type="exec",
            executable=self.executable,
            executable_params=self.executable_params,
            id=f"exec_{self.executable}",
            stdin_source=None,
        )

        self._run_process(p)

    def _final_cleanup(self):
        """
        Final cleanup called in the run function at the very end of processing
        """
        # Nothing to cleanup
        pass



#!/usr/bin/env python

# Copyright 2016 Google, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This application demonstrates how to perform basic operations on blobs
(objects) in a Google Cloud Storage bucket.

For more information, see the README.md under /storage  and the documentation
at https://cloud.google.com/storage/docs.
"""

import argparse
import datetime
from google.cloud import storage


def create_bucket(bucket_name):
    """Creates a new bucket."""
    storage_client = storage.Client()
    bucket = storage_client.create_bucket(bucket_name)
    print("Bucket {} created".format(bucket.name))
    return bucket


def delete_bucket(bucket_name):
    """Deletes a bucket. The bucket must be empty."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    bucket.delete()
    print("Bucket {} deleted".format(bucket.name))


def list_blobs(bucket_name):
    """Lists all the blobs in the bucket."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)

    blobs = bucket.list_blobs()

    for blob in blobs:
        print(blob.name)

    return blobs


def list_blobs_with_prefix(bucket_name, prefix, delimiter=None):
    """Lists all the blobs in the bucket that begin with the prefix.

    This can be used to list all blobs in a "folder", e.g. "public/".

    The delimiter argument can be used to restrict the results to only the
    "files" in the given "folder". Without the delimiter, the entire tree under
    the prefix is returned. For example, given these blobs:

        /a/1.txt
        /a/b/2.txt

    If you just specify prefix = '/a', you'll get back:

        /a/1.txt
        /a/b/2.txt

    However, if you specify prefix='/a' and delimiter='/', you'll get back:

        /a/1.txt

    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)

    blobs = bucket.list_blobs(prefix=prefix, delimiter=delimiter)

    print("Blobs:")
    for blob in blobs:
        print(blob.name)

    if delimiter:
        print("Prefixes:")
        for prefix in blobs.prefixes:
            print(prefix)

    return blobs


def upload_blob(bucket_name, source_file_name, destination_blob_name):
    """Uploads a file to the bucket."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    blob.upload_from_filename(source_file_name)

    print(
        "File {} uploaded to {}.".format(
            source_file_name, destination_blob_name
        )
    )


def download_blob(bucket_name, source_blob_name, destination_file_name):
    """Downloads a blob from the bucket."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(source_blob_name)

    blob.download_to_filename(destination_file_name)

    print(
        "Blob {} downloaded to {}.".format(
            source_blob_name, destination_file_name
        )
    )


def delete_blob(bucket_name, blob_name):
    """Deletes a blob from the bucket."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    blob.delete()

    print("Blob {} deleted.".format(blob_name))


def blob_metadata(bucket_name, blob_name):
    """Prints out a blob's metadata."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.get_blob(blob_name)

    print("Blob: {}".format(blob.name))
    print("Bucket: {}".format(blob.bucket.name))
    print("Storage class: {}".format(blob.storage_class))
    print("ID: {}".format(blob.id))
    print("Size: {} bytes".format(blob.size))
    print("Updated: {}".format(blob.updated))
    print("Generation: {}".format(blob.generation))
    print("Metageneration: {}".format(blob.metageneration))
    print("Etag: {}".format(blob.etag))
    print("Owner: {}".format(blob.owner))
    print("Component count: {}".format(blob.component_count))
    print("Crc32c: {}".format(blob.crc32c))
    print("md5_hash: {}".format(blob.md5_hash))
    print("Cache-control: {}".format(blob.cache_control))
    print("Content-type: {}".format(blob.content_type))
    print("Content-disposition: {}".format(blob.content_disposition))
    print("Content-encoding: {}".format(blob.content_encoding))
    print("Content-language: {}".format(blob.content_language))
    print("Metadata: {}".format(blob.metadata))


def make_blob_public(bucket_name, blob_name):
    """Makes a blob publicly accessible."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    blob.make_public()

    print(
        "Blob {} is publicly accessible at {}".format(
            blob.name, blob.public_url
        )
    )


def generate_signed_url(bucket_name, blob_name):
    """Generates a signed URL for a blob.

    Note that this method requires a service account key file. You can not use
    this if you are using Application Default Credentials from Google Compute
    Engine or from the Google Cloud SDK.
    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    url = blob.generate_signed_url(
        # This URL is valid for 1 hour
        expiration=datetime.timedelta(hours=1),
        # Allow GET requests using this URL.
        method="GET",
    )

    print("The signed url for {} is {}".format(blob.name, url))

    return url


def rename_blob(bucket_name, blob_name, new_name):
    """Renames a blob."""
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    new_blob = bucket.rename_blob(blob, new_name)

    print("Blob {} has been renamed to {}".format(blob.name, new_blob.name))


def copy_blob(bucket_name, blob_name, new_bucket_name, new_blob_name):
    """Copies a blob from one bucket to another with a new name."""
    storage_client = storage.Client()
    source_bucket = storage_client.get_bucket(bucket_name)
    source_blob = source_bucket.blob(blob_name)
    destination_bucket = storage_client.get_bucket(new_bucket_name)

    new_blob = source_bucket.copy_blob(
        source_blob, destination_bucket, new_blob_name
    )

    print(
        "Blob {} in bucket {} copied to blob {} in bucket {}.".format(
            source_blob.name,
            source_bucket.name,
            new_blob.name,
            destination_bucket.name,
        )
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("bucket_name", help="Your cloud storage bucket.")

    subparsers = parser.add_subparsers(dest="command")
    subparsers.add_parser("create-bucket", help=create_bucket.__doc__)
    subparsers.add_parser("delete-bucket", help=delete_bucket.__doc__)
    subparsers.add_parser("list", help=list_blobs.__doc__)

    list_with_prefix_parser = subparsers.add_parser(
        "list-with-prefix", help=list_blobs_with_prefix.__doc__
    )
    list_with_prefix_parser.add_argument("prefix")
    list_with_prefix_parser.add_argument("--delimiter", default=None)

    upload_parser = subparsers.add_parser("upload", help=upload_blob.__doc__)
    upload_parser.add_argument("source_file_name")
    upload_parser.add_argument("destination_blob_name")

    download_parser = subparsers.add_parser(
        "download", help=download_blob.__doc__
    )
    download_parser.add_argument("source_blob_name")
    download_parser.add_argument("destination_file_name")

    delete_parser = subparsers.add_parser("delete", help=delete_blob.__doc__)
    delete_parser.add_argument("blob_name")

    metadata_parser = subparsers.add_parser(
        "metadata", help=blob_metadata.__doc__
    )
    metadata_parser.add_argument("blob_name")

    make_public_parser = subparsers.add_parser(
        "make-public", help=make_blob_public.__doc__
    )
    make_public_parser.add_argument("blob_name")

    signed_url_parser = subparsers.add_parser(
        "signed-url", help=generate_signed_url.__doc__
    )
    signed_url_parser.add_argument("blob_name")

    rename_parser = subparsers.add_parser("rename", help=rename_blob.__doc__)
    rename_parser.add_argument("blob_name")
    rename_parser.add_argument("new_name")

    copy_parser = subparsers.add_parser("copy", help=rename_blob.__doc__)
    copy_parser.add_argument("blob_name")
    copy_parser.add_argument("new_bucket_name")
    copy_parser.add_argument("new_blob_name")

    args = parser.parse_args()

    if args.command == "create-bucket":
        create_bucket(args.bucket_name)
    elif args.command == "delete-bucket":
        delete_bucket(args.bucket_name)
    elif args.command == "list":
        list_blobs(args.bucket_name)
    elif args.command == "list-with-prefix":
        list_blobs_with_prefix(args.bucket_name, args.prefix, args.delimiter)
    elif args.command == "upload":
        upload_blob(
            args.bucket_name, args.source_file_name, args.destination_blob_name
        )
    elif args.command == "download":
        download_blob(
            args.bucket_name, args.source_blob_name, args.destination_file_name
        )
    elif args.command == "delete":
        delete_blob(args.bucket_name, args.blob_name)
    elif args.command == "metadata":
        blob_metadata(args.bucket_name, args.blob_name)
    elif args.command == "make-public":
        make_blob_public(args.bucket_name, args.blob_name)
    elif args.command == "signed-url":
        generate_signed_url(args.bucket_name, args.blob_name)
    elif args.command == "rename":
        rename_blob(args.bucket_name, args.blob_name, args.new_name)
    elif args.command == "copy":
        copy_blob(
            args.bucket_name,
            args.blob_name,
            args.new_bucket_name,
            args.new_blob_name,
        )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Process queue implementation using multiprocessing, Queue() and Thread.

The process queue is responsible to run all requests in actinia that
require to execute GRASS GIS processes or UNIX processes to create a response.

The process queue supports logging of the stderr output of the executed
processes into a rotating logfile and fluent server.
"""

import pickle
import time
from datetime import datetime
import queue as standard_queue
from multiprocessing import Process, Queue
from threading import Thread, Lock
import logging
import atexit
from actinia_core.core.resources_logger import ResourceLogger
from actinia_core.core.logging_interface import log


has_fluent = False

try:
    from fluent import handler

    has_fluent = True
except Exception:
    print("Fluent is not available")
    has_fluent = False


__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-present, Sören Gebbert and mundialis GmbH & Co. KG"
)

process_queue = Queue()
process_queue_manager = None


def create_process_queue(config, use_logger=True):
    """Create the process queue that will start all processes in a separate
    process. It uses a multiprocessing.Queue() to receive Processes
    (function and arguments)

    The process queue can only be started once

    Args:
        config: The global configuration
        use_logger: Use the rotating file logger and fluent for stderr logging
                    of the processes
    """
    global process_queue_manager

    if process_queue_manager is None:
        p = Process(
            target=start_process_queue_manager,
            args=(config, process_queue, use_logger),
        )
        p.start()
        process_queue_manager = p


def enqueue_job(timeout, func, *args):
    """Put the provided function and arguments in the process queue

    Args:
        timeout: The timeout of the process, if the timeout is exceeded by the
                 process it will be killed
        func: The function to call from the subprocess
        *args: The function arguments, the first argument must be the
               RessourceDataContainer
    """
    process_queue.put((func, timeout, args))

    # # for debugging in ephemeral_processing.py (see also grass_init.py)
    # # only uncomment ONE of the following endpoints:

    # # for '/locations/<string:location_name>/processing_async'
    # from ...rest.ephemeral_processing import \
    #     AsyncEphemeralResource
    # from ...processing.common.ephemeral_processing import start_job
    # from ...processing.actinia_processing.ephemeral_processing \
    #      import EphemeralProcessing
    # processing = EphemeralProcessing(*args)
    # processing.run()

    # # for '/locations/<string:location_name>/processing_async_export'
    # from ...processing.actinia_processing.ephemeral.\
    #     ephemeral_processing_with_export import EphemeralProcessingWithExport
    # processing = EphemeralProcessingWithExport(*args)
    # processing.run()

    # # for /locations/{location_name}/mapsets/{mapset_name}/processing_async
    # from ...processing.actinia_processing.ephemeral.persistent_processing \
    #     import PersistentProcessing
    # processing = PersistentProcessing(*args)
    # processing.run()


def stop_process_queue():
    """Destroy the process queue and terminate all running and enqueued jobs"""
    global process_queue_manager
    # Send stop to the queue
    process_queue.put("STOP")
    # Wait for all joining processes
    if process_queue_manager:
        # print("Waited for process_queue_manager")
        process_queue_manager.join(3)
        # print("Terminate process_queue_manager")
        process_queue_manager.terminate()
    process_queue_manager = None


# Register the stop_process_queue in the exit handler
atexit.register(stop_process_queue)


class EnqueuedProcess(object):
    """The class that takes care of the handling of a single process. It
    provides support for timeout check, exit status check and resource
    termination commits.
    It implements methods to start and gently terminate processes, so that the
    resource database will be updated with the current state.

    - timeout check -- Check if a waiting process exceedes its timeout for
                       waiting to be run and terminate it.
                       A resource update will be send to the resource database.
    - exits status check -- Check if the exit status of the process was 0, if
                            not check if the resource database acknowledged
                            this with a termination or error message, if not
                            send a resource update
    - termination commits - Terminate the process and send an update to the
                            resource database about the termination
    """

    def __init__(self, func, timeout, resource_logger, args):
        self.process = Process(target=func, args=args)
        self.timeout = timeout
        self.config = args[0].config
        self.resource_id = args[0].resource_id
        self.iteration = args[0].iteration
        self.user_id = args[0].user_id
        self.api_info = args[0].api_info
        self.resource_logger = resource_logger
        self.init_time = time.time()

        self.started = False

    def __del__(self):
        pass
        # print("Process deleted", self.resource_id)

    def start(self):
        """Start the process

        :return:
        """
        # print("Start job: ", self.api_info)
        self.started = True
        self.process.start()

    def terminate(self, status, message):
        """Terminate the process

        Send a termination response to the resource database.

        Args:
            status: The status why termination was requested
            message: The message why the process was terminated by the server
                     (timeout, server shutdown, ...)
        """
        # print("Terminate process with message: ", message)

        if self.process.is_alive():
            self.process.terminate()

        self._send_resource_update(status=status, message=message)

    def is_alive(self):
        return self.process.is_alive()

    def exitcode(self):
        return self.process.exitcode

    def check_timeout(self):
        """
        Check if the process waited longer for running then the timeout that
        was set

        Terminate the process if the timeout limit was exceeded.

        Returns:
             False if within timeout, True if the process terminated itself
        """
        # print("Check timeout for process: ", self.api_info)
        if self.started is False:
            current_time = time.time()
            diff = current_time - self.init_time
            if self.timeout < diff:
                self.terminate(
                    status="timeout",
                    message="Processes exceeded timeout (%i) in "
                    "waiting queue and was terminated." % self.timeout,
                )
                return True

        return False

    def check_exit(self):
        """Check the exitcode, if a non-zero exit code was received then
        send an update to the resource logger that something strange happened.
        Send only if the status of the resource is not "error", "terminated"
        or "timeout".

        """
        if self.process.exitcode is not None and self.process.exitcode != 0:
            # Check if the process noticed the error already
            response_data = self.resource_logger.get(
                self.user_id, self.resource_id, self.iteration
            )

            if response_data is not None:
                http_code, response_model = pickle.loads(response_data)
                if (
                    response_model["status"] != "error"
                    and response_model["status"] != "terminated"
                    and response_model["status"] != "timeout"
                ):
                    message = (
                        "The process unexpectedly terminated with exit code %i"
                        % self.process.exitcode
                    )
                    self._send_resource_update(
                        status="error",
                        message=message,
                        response_data=response_data,
                    )

    def _send_resource_update(self, status, message, response_data=None):
        """
        Send a response to the resource logger about the current resource state

        Args:
            status: The status that should be set (terminated)
            message: The message
        """
        # print("Send resource update status: ", status, " message: ", message)
        # Get the latest response and use it as template for the resource
        # update
        if response_data is None:
            response_data = self.resource_logger.get(
                self.user_id, self.resource_id, self.iteration
            )

        # Send the termination response
        if response_data is not None:
            http_code, response_model = pickle.loads(response_data)
            # print("Resource", http_code, response_model)
            response_model["status"] = status
            response_model["message"] = (
                "The process was terminated by the server: %s" % message
            )
            orig_time = response_model["accept_timestamp"]
            response_model["timestamp"] = time.time()
            response_model["datetime"] = str(datetime.now())
            response_model["time_delta"] = (
                response_model["timestamp"] - orig_time
            )

            document = pickle.dumps([http_code, response_model])

            self.resource_logger.commit(
                user_id=self.user_id,
                resource_id=self.resource_id,
                iteration=self.iteration,
                document=document,
                expiration=self.config.REDIS_RESOURCE_EXPIRE_TIME,
            )


def queue_watcher(queue, data_set, lock):
    """This function runs in a separate thread to check the queue

    Args:
        queue: The queue to check
        data_set: The set to add the received data
        lock: The lock to be used for locking the set access
    """

    while True:
        try:
            # print("Check for new data in queue")
            data = queue.get(block=True)
            lock.acquire_lock()
            data_set.add(data)
            # print("Add data to set", len(data_set))
            lock.release_lock()
        except standard_queue.Empty:
            pass


def start_process_queue_manager(config, queue, use_logger):
    """
    The process queue manager that runs the infinite loop for worker creation

    - This function creates the stderr logger if requested
    - It listen to a queue in an infinite loop:
        - The queue is watched in a separate thread so that no data get lost
        - Check if new processes are in the data set() that is filled by the
          queue watcher
        - Check the timeout of waiting processes
        - Enqueue and start new processes
        - Remove finished processes or processes that exceeded their waiting
          timeout
        - Stop the queue and exit all running processes if the "STOP" isgnal
          was send via Queue()

    Args:
        config: The global config
        queue: The multiprocessing.Queue() object that should be listened to
        use_logger: Create logifle and fluent logger to log the stderr of the
                    processes
    """
    global finished_procs

    data_set = set()
    lock = Lock()
    # Start the thread that permanently listens to the queue
    queue_thread = Thread(target=queue_watcher, args=(queue, data_set, lock))
    queue_thread.start()

    running_procs = set()
    waiting_processes = set()

    fluent_sender = None
    # Fluentd hack to work in a multiprocessing environment
    try:
        from fluent import sender

        fluent_sender = sender.FluentSender(
            "actinia_process_logger",
            host=config.LOG_FLUENT_HOST,
            port=config.LOG_FLUENT_PORT,
        )
    except Exception:
        pass
    # We need the resource logger to send updates to the resource database
    kwargs = dict()
    kwargs["host"] = config.REDIS_SERVER_URL
    kwargs["port"] = config.REDIS_SERVER_PORT
    if config.REDIS_SERVER_PW and config.REDIS_SERVER_PW is not None:
        kwargs["password"] = config.REDIS_SERVER_PW
    # Seems to log only on rare occasions: when a job waits too long in the
    # queue and terminates itself.
    resource_logger = ResourceLogger(**kwargs, fluent_sender=fluent_sender)
    del kwargs

    count = 0
    try:
        while True:
            # Get the process data from the set that is filled in the queue
            # thread
            data = None
            lock.acquire_lock()
            if len(data_set) > 0:
                # print("Jobs from queue: ", len(data_set))
                data = data_set.pop()
            lock.release_lock()

            if data is not None:
                # Stop all (running and waiting) processes if the STOP command
                # was detected and leave the loop
                if "STOP" in data:
                    for enqproc in running_procs:
                        enqproc.terminate(
                            status="error",
                            message="Running process was terminated by server "
                            "shutdown.",
                        )
                    for enqproc in waiting_processes:
                        enqproc.terminate(
                            status="error",
                            message="Waiting process was terminated by server "
                            "shutdown.",
                        )
                    del queue_thread
                    queue.close()
                    # print("Exit loop")
                    exit(0)
                # Enqueue a new process
                elif len(data) == 3:
                    func, timeout, args = data
                    log.info("Enqueue process: %s", args[0].api_info)
                    enqproc = EnqueuedProcess(
                        func=func,
                        timeout=timeout,
                        resource_logger=resource_logger,
                        args=args,
                    )
                    waiting_processes.add(enqproc)

            if len(running_procs) < config.NUMBER_OF_WORKERS:
                if len(waiting_processes) > 0:
                    enqproc = waiting_processes.pop()
                    running_procs.add(enqproc)
                    log.info("Run process: %s", enqproc.api_info)
                    enqproc.start()

            # Purge processes that are finished or exceeded their timeout each
            # 40th loop
            if count % 10 == 0:
                count = 0
                procs_to_remove = []
                # purge processes that has been finished
                for enqproc in running_procs:
                    if enqproc.started is True and enqproc.is_alive() is False:
                        # Check if the process finished with an error and send
                        # a resource update if required
                        enqproc.check_exit()
                        procs_to_remove.append(enqproc)
                for enqproc in procs_to_remove:
                    running_procs.remove(enqproc)

                procs_to_remove = []
                # purge processes that have exceeded their timeout for waiting
                for enqproc in waiting_processes:
                    check = enqproc.check_timeout()
                    if check is True:
                        procs_to_remove.append(enqproc)
                for enqproc in procs_to_remove:
                    waiting_processes.remove(enqproc)

            time.sleep(0.05)
            count += 1
    except Exception:
        raise
    finally:
        queue.close()
